/*****************************************************************************
 * File: MITeSignal.cpp
 * Author: Mark Blum (mark@media.mit.edu), MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * 
 * Copyright (C) 2005 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************
 * This Program is used to calibrate accelerometers. Input are one or two
 * (hardcoded) acceleration signals, output is a textfile containing calibra-
 * tion values.
 * static method: not really used because unsuitable
 * dynamic method: We first determine the 1g-equivalent values for all axis. 
 * For that each accelerometer has to be rotated in each of its six extreme 
 * orientations and held still for a few seconds. Two rolling averages are 
 * computed per axis. One can only increase, the other only decrease and the 
 * learning factor is indirectly proportional to the variance over a 128-sample 
 * window. Also, the values are median filtered, which further reduces the risk 
 * of distorting the calibration by shaking. The average values for each axis 
 * corresponding to zero g are finally the mean of both 1g-equivalent values. 
 * The difference, divided by two, is scale corresponding to 1g.
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <enchantment/signals.h>

#define LEARNTIME 30
#define MAXVAR 100.0
#define RING_BUFF_SIZE 128

void PrintUsage()
{
    printf("Usage: MITeCalibrate -s [unl1] [unl2] -d [unl1] [unl2] -o [output File] -r [sampling rate] -t [time]\n");
    printf("                                                                                              \n");
    printf("       -s are static calibrations (a rolling average is computed on the current orrientation) \n");
    printf("          --> keep MITe still, ideal for the one on the hip                                   \n");
    printf("       -d are dynamic calibrations (a seperate rolling average is computed for the maximal    \n");
    printf("          and minimal values, then the arithmetic middle is chosen                            \n");
    printf("          --> rotate MITe softly around all axes, ideal for the one on the wrist              \n");
    printf("       -o is the path and filename-prefix for the output files:                               \n");
    printf("          '_MITe_X.txt' will be appended for each UNL                                         \n");
    printf("          '_MITes.txt' will be appended for one file containing all calibrations.             \n");
    printf("       -r Sampling rate in herz                                                               \n");
    printf("       -t calibration time in seconds                                                         \n");
    printf("       Currently, not more that two (2) MITes can be calibrated at once                       \n");
    printf("                                                                                              \n");
    printf("       Defaults:  -s [/signals/MITE/MITe_5]\n");
    printf("                  -d [/signals/MITE/MITe_9]\n");
    printf("                  -o [Config]\n");
    printf("                  -r [90]\n");
    printf("                  -t [30]\n");
    printf("\n");
}

int compare (const void * a, const void * b)
{
    return ( *(short*)a - *(short*)b );
}

inline float CalcVar(short *inVec, int windowSize) {
    int i, sumMean = 0;
    float mean, sumVar = 0;
    for (i=0; i < windowSize; i++){
        sumMean += inVec[i];
    }
    mean = (float)sumMean/((float)windowSize);
    
    for (i=0; i < windowSize; i++){
        sumVar += inVec[i]*inVec[i];
    }
    return sumVar/((float)windowSize) - mean*mean;            // var(X) = E(X - u)^2 = E(X^2) - u^2
}

int main(int argc, char *argv[]) 
{
    FILE *fid1, *fid2, *fidAll;
    char *outFile1, *outFile2, *outFileAll;
    Signal *inSignal1;
    Signal *inSignal2;
    struct timeval inTimestamp;
    short inBuffer[3];
    float high1[] = {200,200,200};
    float high2[] = {200,200,200};
    float low1[] = {300,300,300};
    float low2[] = {300,300,300};
    float medianHigh[] = {250,250,250};
    float medianLow[] = {250,250,250};
    float currentVariance[] = {0,0,0};
    float learnSpeed[] = {0,0,0};
    float average1[] = {250,250,250};
    float average2[] = {250,250,250};
    float one_G1[] = {60,60,60};
    float one_G2[] = {60,60,60};
    float accumulatedLearning1[] = {0,0,0,0,0,0};
    float accumulatedLearning2[] = {0,0,0,0,0,0};
    int i,k,index1, index2, num_UNLs_read = 0;
    short valsX1[RING_BUFF_SIZE], valsY1[RING_BUFF_SIZE], valsZ1[RING_BUFF_SIZE];
    short valsX2[RING_BUFF_SIZE], valsY2[RING_BUFF_SIZE], valsZ2[RING_BUFF_SIZE];
    short valsTemp[RING_BUFF_SIZE];
    int finishedCalibrating1, finishedCalibrating2;
    
    //these are the variables defining the default behaviour
    // idea: MITe 1 on hip, static, MITe 2 on wrist, dynamic
    char *inUNL1 = "/signals/MITe/MITe_5";
    char *inUNL2 = "/signals/MITe/MITe_9";
    char *outputFilePrefix = "Config";
    int twoSensors = 1;
    int sampleRate = 90;      //[Herz]
    int calibrationTime = 30;  //[seconds]
    int dynamic[] = {0,1};    //{'static', 'dynamic'}
    
    
    ///// Parse input arguments ///////////////////////////////////////////

    num_UNLs_read = 0;
    i = 1;
    while ( i < argc )
    {
        if (argv[i][0] == '-' )
        {
            i++;
            if ( i == argc )
            {
                PrintUsage();
                return -1;
            }
            switch ( argv[i-1][1] )
            {
                case 'o': {
                    outputFilePrefix = argv[i];
                    break;
                }
                case 'r': {
                    sampleRate = atoi(argv[i]);
                    break;
                }
                case 't': {
                    calibrationTime = atoi(argv[i]);
                    break;
                }
                case 's': {
                    switch (num_UNLs_read){
                        case 0: {
                            inUNL1 = argv[i];
                            dynamic[num_UNLs_read] = 0;
                            num_UNLs_read++;
                            twoSensors = 0;
                            if ( i < argc-1 ){
                                if (argv[i+1][0] != '-' ){
                                    inUNL2 = argv[++i];
                                    dynamic[num_UNLs_read] = 0;
                                    num_UNLs_read++;
                                    twoSensors = 1;
                                }
                            }
                            break;
                        }
                        case 1: {
                            inUNL2 = argv[i];
                            dynamic[num_UNLs_read] = 0;
                            num_UNLs_read++;
                            twoSensors = 1;
                            break;
                        }
                        default: 
                            printf("Currently, not more that two (2) MITes can be calibrated at once!\n");
                            PrintUsage(); 
                            return -1;
                    }
                    break;
                }
                case 'd': {
                    switch (num_UNLs_read){
                        case 0: {
                            inUNL1 = argv[i];
                            dynamic[num_UNLs_read] = 1;
                            num_UNLs_read++;
                            twoSensors = 0;
                            if ( i < argc-1 ){
                                if (argv[i+1][0] != '-' ){
                                    inUNL2 = argv[++i];
                                    dynamic[num_UNLs_read] = 1;
                                    num_UNLs_read++;
                                    twoSensors = 1;
                                }
                            }
                            break;
                        }
                        case 1: {
                            inUNL2 = argv[i];
                            dynamic[num_UNLs_read] = 1;
                            num_UNLs_read++;
                            twoSensors = 1;
                            break;
                        }
                        default: 
                            printf("Currently, not more that two (2) MITes can be calibrated at once!\n");
                            PrintUsage(); 
                            return -1;
                    }
                    break;
                }
                default: PrintUsage(); return -1;
            }
            i++;
        }
        else 
        { 
            PrintUsage();
            return -1;
        }
    }  ///// end parsing input arguments
    
    if (twoSensors){
        printf("MITeCalibrate: inUNL: %s (%s)  inUNL2: %s (%s)  outputfile: %s\n", inUNL1,(dynamic[0]==0) ? "static" : "dynamic",inUNL2,(dynamic[1]==0) ? "static" : "dynamic",outputFilePrefix);
    } else {
        printf("MITeCalibrate: inUNL: %s (%s)  outputfile: %s\n", inUNL1,(dynamic[0]==0) ? "static" : "dynamic",outputFilePrefix);
    }
    
    ///// setup input signals and output file /////////////////////////////
    
    
    inSignal1 = NewSignalUNL(inUNL1);
    if(inSignal1 == NULL) {
        printf("MITeCalibrate: Error allocating input signal %s\n",inUNL1);
        exit(1);
    }
    if (!SignalConnectWait(inSignal1)){
        printf("MITeCalibrate: Error connecting to %s\n",inUNL1);
        exit(1);
    }
    if (SignalGetType(inSignal1) != T_INT16 || SignalGetLength(inSignal1, 1) != 3)
    {
        printf("in signal type error: type: #%d; columns: %d\n",
               SignalGetType(inSignal1), SignalGetLength(inSignal1, 1));
        exit(1);
    }
    outFile1 = (char*) malloc(strlen(outputFilePrefix)+strlen(strrchr(inUNL1,'/'))+8);
    sprintf(outFile1,"%s_%s.txt",outputFilePrefix,strrchr(inUNL1,'/')+1);
    fid1 = fopen(outFile1,"w");
    if (fid1 == NULL) {
        printf("MITeCalibrate: Error, could not create file %s\n",outFile1);
        exit(1);
    }
    else {
        fclose(fid1);
    }
    outFileAll = (char*) malloc(strlen(outputFilePrefix)+13);
    sprintf(outFileAll,"%s_MITes.txt",outputFilePrefix);
    fidAll = fopen(outFileAll,"w");
    if (fidAll == NULL) {
        printf("MITeCalibrate: Error, could not create file %s\n",outFileAll);
        exit(1);
    }
    else {
        fclose(fidAll);
    }
    
    if (twoSensors)
    {
        inSignal2 = NewSignalUNL(inUNL2);
        if(inSignal2 == NULL) {
            printf("AccelFeature: Error allocating inputsignal%s\n",inUNL2);
            exit(1);
        }
        if (!SignalConnectWait(inSignal2)){
            printf("AccelFeature: Error connecting to %s\n",inUNL2);
            exit(1);
        }
        if (SignalGetType(inSignal2) != T_INT16 || SignalGetLength(inSignal2, 1) != 3)
        {
            printf("in signal type error: type: #%d; columns: %d\n",
                   SignalGetType(inSignal2), SignalGetLength(inSignal2, 1));
            exit(1);
        }
    
        outFile2 = (char*) malloc(strlen(outputFilePrefix)+strlen(strrchr(inUNL2,'/'))+8);
        sprintf(outFile2,"%s_%s.txt",outputFilePrefix,strrchr(inUNL2,'/')+1);
        fid2 = fopen(outFile2,"w");
        if (fid2 == NULL) {
            printf("MITeCalibrate: Error, could not create file %s\n",outFile2);
            exit(1);
        }
        else {
            fclose(fid2);
        }
    }
                      
    printf("MITeCalibrate: Starting calibration...\n");
    printf("               Remember: for static calibration, hold the MITe still in the position it is worn.\n");
    printf("               For dynamic calibration, rotate the MITe gently around all three axes.\n");
        
    for(i=0; i<RING_BUFF_SIZE; i++) {
        valsX1[i] = 250;        valsY1[i] = 250;        valsZ1[i] = 250;
        valsX2[i] = 250;        valsY2[i] = 250;        valsZ2[i] = 250;
    }
    
    index1 = 0;
    index2 = 0;
            
    //// Main loop ////////////////////////////////////////////////////////////////////////////////////
    
    for (k=0; k<sampleRate*calibrationTime; k++){
        
        SignalRecv(inSignal1, &inTimestamp, (void *) inBuffer);
        
        valsX1[index1] = inBuffer[0];
        valsY1[index1] = inBuffer[1];
        valsZ1[index1] = inBuffer[2];
        index1 = (index1 + 1) % RING_BUFF_SIZE;
            
        //compute the median and variance of each axis over a window
        memcpy(valsTemp,valsX1,RING_BUFF_SIZE*sizeof(short));
        qsort(valsTemp, RING_BUFF_SIZE, sizeof(short), compare);
        medianHigh[0] = valsTemp[(int)(RING_BUFF_SIZE/2.1)];
        medianLow[0] = valsTemp[(int)(RING_BUFF_SIZE/1.9)];
        currentVariance[0] = CalcVar(valsTemp,RING_BUFF_SIZE);
            
        memcpy(valsTemp,valsY1,RING_BUFF_SIZE*sizeof(short));
        qsort(valsTemp, RING_BUFF_SIZE, sizeof(short), compare);
        medianHigh[1] = valsTemp[(int)(RING_BUFF_SIZE/2.1)];
        medianLow[1] = valsTemp[(int)(RING_BUFF_SIZE/1.9)];
        currentVariance[1] = CalcVar(valsTemp,RING_BUFF_SIZE);
            
        memcpy(valsTemp,valsZ1,RING_BUFF_SIZE*sizeof(short));
        qsort(valsTemp, RING_BUFF_SIZE, 2, compare);
        medianHigh[2] = valsTemp[(int)(RING_BUFF_SIZE/2.1)];
        medianLow[2] = valsTemp[(int)(RING_BUFF_SIZE/1.9)];
        currentVariance[2] = CalcVar(valsTemp,RING_BUFF_SIZE);
              
//             if (0) {
//                 printf("\n valsTempSort: ");
//                 for (i=0; i<RING_BUFF_SIZE; i++) {
//                     printf("%d ",valsTemp[i]);
//                 }
//                 printf("\n Variance: ");
//                 for (i=0; i<3; i++){
//                     printf("%f ", currentVariance[i]);
//                 } 
//                 printf("\n MedianHigh: ");
//                 for (i=0; i<3; i++){
//                     printf("%f ", medianHigh[i]);
//                 } 
//                 printf("\n MedianLow: ");
//                 for (i=0; i<3; i++){
//                     printf("%f ", medianLow[i]);
//                 } 
//                 printf("\n");
//                 exit(0);
//             }
              
        for (i=0; i<3; i++){
            // Var = 0 --> learnSpeed[i] = 1 ; Var = MAXVAR --> learnSpeed[i] = 0
            learnSpeed[i] = -1/MAXVAR*currentVariance[i] + 1;
            learnSpeed[i] = (learnSpeed[i] > 1) ? 1 : learnSpeed[i]; 
            learnSpeed[i] = (learnSpeed[i] < 0) ? 0 : learnSpeed[i]; 
        }
        
        if (dynamic[0]) {
            
            for (i=0; i<3; i++){
                if (medianHigh[i] > high1[i]){
                    high1[i] = learnSpeed[i]*medianHigh[i]+(1-learnSpeed[i])*high1[i];    
                    accumulatedLearning1[i] += learnSpeed[i];
                }
                if (medianLow[i] < low1[i]){
                    low1[i] = learnSpeed[i]*medianLow[i]+(1-learnSpeed[i])*low1[i];
                    accumulatedLearning1[i+3] += learnSpeed[i];
                }
                average1[i] = (high1[i]+low1[i]) / 2;
                one_G1[i] = high1[i] - average1[i];
            }
        }
        else // static
        {
            for (i=0; i<3; i++){
                average1[i] = learnSpeed[i]*inBuffer[i]+(1-learnSpeed[i])*average1[i];
                accumulatedLearning1[i] += learnSpeed[i];
                accumulatedLearning1[i+3] = accumulatedLearning1[i];
                if (twoSensors && dynamic[1]){  
                    // special case (actually defualt case) where we have one static and one dynamic
                    // we can guess the one_G range by taking the average of the dynamic measurment
                    // thus assuming that both MITes have about the same resolution (integers per G)
                    one_G1[i] = (one_G2[0]+one_G2[1]+one_G2[2])/3;
                } 
                else{
                    one_G1[i] = 60;  // this is a manual measurment done previously
                }
            }
        }
        
        finishedCalibrating1 = 1;
        for (i=0; i<3; i++){
            finishedCalibrating1 = finishedCalibrating1 && (accumulatedLearning1[i] > LEARNTIME) && (accumulatedLearning1[i+3] > LEARNTIME);
        }
            
        if ( k % sampleRate == 0){       // write current results to file once per second
            fid1 = fopen(outFile1,"w");
            if (fid1 == NULL) {
                printf("MITeCalibrate: Error, could not create file %s\n",outFile1);
                exit(1);
            }
//             for (i=0; i<3; i++){
//                 printf("%f ", currentVariance[i]);
//             } printf("\n");
            fprintf(fid1,"%4.1f %4.1f %4.1f %4.1f %4.1f %4.1f\n",average1[0],average1[1],average1[2],one_G1[0],one_G1[1],one_G1[2]);
            fclose(fid1);
            fprintf(stdout,"%s: %4.1f %4.1f %4.1f %4.1f %4.1f %4.1f\n",strrchr(inUNL1,'/')+1,
                    average1[0],average1[1],average1[2],one_G1[0],one_G1[1],one_G1[2]);
            fidAll = fopen(outFileAll,"w");
            if (fidAll == NULL) {
                printf("MITeCalibrate: Error, could not create file %s\n",outFileAll);
                exit(1);
            }
            fprintf(fidAll,"%4.1f %4.1f %4.1f %4.1f %4.1f %4.1f\n",average1[0],average1[1],average1[2],one_G1[0],one_G1[1],one_G1[2]);
            fclose(fidAll);
        }

        ///// second MITe ////////////////////////////////////////////////////////////////////////////////
        //
        // by regularly reading one value from each signal, we assume the samling rate is more or less equal
        // This is ok because for calibration purposes it doesn't matter if we shift a bit in time
        if (twoSensors) {
            SignalRecv(inSignal2, &inTimestamp, (void *) inBuffer);
            
            valsX2[index2] = inBuffer[0];
            valsY2[index2] = inBuffer[1];
            valsZ2[index2] = inBuffer[2];
            
            index2 = (index2 + 1) % RING_BUFF_SIZE;
            
            
                
            //compute the median and variance of each axis over a window
            memcpy(valsTemp,valsX2,RING_BUFF_SIZE*sizeof(short));
            qsort(valsTemp, RING_BUFF_SIZE, sizeof(short), compare);
            medianHigh[0] = valsTemp[(int)(RING_BUFF_SIZE/2.1)];
            medianLow[0] = valsTemp[(int)(RING_BUFF_SIZE/1.9)];
            currentVariance[0] = CalcVar(valsTemp,RING_BUFF_SIZE);
                
            memcpy(valsTemp,valsY2,RING_BUFF_SIZE*sizeof(short));
            qsort(valsTemp, RING_BUFF_SIZE, sizeof(short), compare);
            medianHigh[1] = valsTemp[(int)(RING_BUFF_SIZE/2.1)];
            medianLow[1] = valsTemp[(int)(RING_BUFF_SIZE/1.9)];
            currentVariance[1] = CalcVar(valsTemp,RING_BUFF_SIZE);
                
            memcpy(valsTemp,valsZ2,RING_BUFF_SIZE*sizeof(short));
            qsort(valsTemp, RING_BUFF_SIZE, sizeof(short), compare);
            medianHigh[2] = valsTemp[(int)(RING_BUFF_SIZE/2.1)];
            medianLow[2] = valsTemp[(int)(RING_BUFF_SIZE/1.9)];
            currentVariance[2] = CalcVar(valsTemp,RING_BUFF_SIZE);
                
            for (i=0; i<3; i++){
                // Var = 0 --> learnSpeed[i] = 1 ; Var = MAXVAR --> learnSpeed[i] = 0
                learnSpeed[i] = -1/MAXVAR*currentVariance[i] + 1;
                learnSpeed[i] = learnSpeed[i] > 1 ? 1 : learnSpeed[i]; 
                learnSpeed[i] = learnSpeed[i] < 0 ? 0 : learnSpeed[i]; 
            }
            
            if (dynamic[1]){    
                
                for (i=0; i<3; i++){
                    if (medianHigh[i] > high2[i]){
                        high2[i] = learnSpeed[i]*medianHigh[i]+(1-learnSpeed[i])*high2[i];
                        accumulatedLearning2[i] += learnSpeed[i];
                    }
                    if (medianLow[i] < low2[i]){
                        low2[i] = learnSpeed[i]*medianLow[i]+(1-learnSpeed[i])*low2[i];
                        accumulatedLearning2[i+3] += learnSpeed[i];
                    }
                    average2[i] = (high2[i]+low2[i]) / 2;
                    one_G2[i] = high2[i] - average2[i];
                }
            }
            else // static
            {
                for (i=0; i<3; i++){
                    average2[i] = learnSpeed[i]*inBuffer[i]+(1-learnSpeed[i])*average2[i];
                    accumulatedLearning2[i] += learnSpeed[i];
                    accumulatedLearning2[i+3] = accumulatedLearning2[i];
                    if (twoSensors && dynamic[0]){  
                    // see above, same thing just inversed
                        one_G2[i] = (one_G1[0]+one_G1[1]+one_G1[2])/3;
                    } 
                    else{
                        one_G2[i] = 60;  // this is a manual measurment done previously
                    }
                }
            }
         
            finishedCalibrating2 = 1;
            for (i=0; i<3; i++){
                finishedCalibrating2 = finishedCalibrating2 && (accumulatedLearning2[i] > LEARNTIME) && (accumulatedLearning2[i+3] > LEARNTIME);
            }
               
            if ( k % sampleRate == 0){       // write current results to file once per second
                fid2 = fopen(outFile2,"w");
                if (fid2 == NULL) {
                    printf("MITeCalibrate: Error, could not create file %s\n",outFile2);
                    exit(1);
                }
                fprintf(fid2,"%4.1f %4.1f %4.1f %4.1f %4.1f %4.1f\n",average2[0],average2[1],average2[2],one_G2[0],one_G2[1],one_G2[2]);
                fclose(fid2);
                fprintf(stdout,"%s: %4.1f %4.1f %4.1f %4.1f %4.1f %4.1f\n",strrchr(inUNL2,'/')+1,
                        average2[0],average2[1],average2[2],one_G2[0],one_G2[1],one_G2[2]);
                fidAll = fopen(outFileAll,"a");
                if (fidAll == NULL) {
                    printf("MITeCalibrate: Error, could not append to file %s\n",outFileAll);
                    exit(1);
                }
                fprintf(fidAll,"%4.1f %4.1f %4.1f %4.1f %4.1f %4.1f\n",average2[0],average2[1],average2[2],one_G2[0],one_G2[1],one_G2[2]);
                fclose(fidAll);
            }   
        } // end if(twoSensors)
        
        
        if (finishedCalibrating1 && finishedCalibrating2) {
            printf("done!\n");
            break;
        }
            
    
    } //end main for-loop
 
    printf("Calibration results for %s\n",strrchr(inUNL1,'/')+1);
    printf("Accumulated Learning:");
    for (i=0; i<6; i++){
        printf(" %2.0f",accumulatedLearning1[i]);
    }
    printf("\nResulting Values: ");
    printf("%4.1f %4.1f %4.1f %4.1f %4.1f %4.1f\n",average1[0],average1[1],average1[2],one_G1[0],one_G1[1],one_G1[2]);
    
    if (twoSensors) {
        printf("Calibration results for %s\n",strrchr(inUNL2,'/')+1);
        printf("Accumulated Learning:");
        for (i=0; i<6; i++){
            printf(" %2.0f",accumulatedLearning2[i]);
        }
        printf("\nResulting Values: ");
        printf("%4.1f %4.1f %4.1f %4.1f %4.1f %4.1f\n",average2[0],average2[1],average2[2],one_G2[0],one_G2[1],one_G2[2]);
    }
    
    free(outFile1);
    if (twoSensors) free(outFile2);
    free(outFileAll);
    return 0;
}
