/*****************************************************************************
 * File: SakSignal.cpp
 * Derived from: Sak2v2Signal.cpp
 * Author: Jonathan Gips, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Adapted from HoarderZaurusSignal.cpp, AccelZaurusSignal.cpp
 * Copyright (C) 2005 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <signal.h>
#include <math.h>

#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>

extern "C" {

#include <enchantment/signals.h>
#include <enchantment/unl.h>
#include "sak2v2_serial.h"
}

const int PRINT_TIMING = 0;
const int  kFailedReadsBeforeEcho = 2;
const int kLoopPause = 2;

typedef struct {
  bool valid;
  Signal *sig;
  char *feature; //mallocd
  struct timeval sampTime;
  char *name;
  int reqSampRate;
  sak2v2_devName devName;
  bool freshSample;
} ChannelRec;

// Prototypes...
void ProcessCommandLine(char *argv[], int argc, 
			char **server, char **user, bool *printData);
void SetupSignalHandlers();
void SigHandler(int sig);
void SetupEnchantSignals(char *server, char *user);
int ReadChannels();  
void PrintChannels();
void SendSignals();
void UpdateSignalConnections(bool *bListeners);
int ConfigureSak2Devices();
void CloseDownSak2();

volatile int runflag = 1;
volatile int killflag = 0;

ChannelRec channelRecs[SAK2V2_NUM_DEVICES];

int main(int argc, char *argv[]) 
{ 
  char *server = NULL, *user = NULL;
  bool bListeners(false), bPrintData(false), bReadOnce(false);
  
  if(0)
    {
      sak2v2_init_serial();  

      while(1)
	{
	  int result = sak2v2_echo(35);
	  printf("echo: %d\n", result);

       
	}
    }
  
  ProcessCommandLine(argv, argc, &server, &user, &bPrintData);
  
  SetupSignalHandlers();
  
  SetupEnchantSignals(server, user);

  
  while(runflag)
    {

      if(SAK_DEBUG) printf("Starting loop: Pause %d seconds.\n", kLoopPause);
      sleep(kLoopPause);

      sak2v2_init_serial();  
      
      if(!sak2v2_echo(8))
	{
	  printf("Sak2 failed to return echo. Resetting.\n");
	  CloseDownSak2();
	  continue;
	}
      
      if(!sak2v2_clear_configuration())
	{
	  printf("Sak2 failed to clear configuration.\n");
	  continue;
	}

      if(!ConfigureSak2Devices())
	{
	  printf("Failed to configure sak2.  Resetting.\n");
	  CloseDownSak2();
	  continue;
	}
	  
      if(!sak2v2_verify_device_configurations())
	{
	  printf("Failed to verify sak2 configuration.  Resetting.\n");
	  CloseDownSak2();
	  continue;
	}
      
      if(!sak2v2_start_streaming())
	{
	  printf("Failed to start sak2 streaming.  Resetting.\n");
	  CloseDownSak2();
	  continue;
	}
      
      if(SAK_DEBUG) printf("main: starting read cycle\n");

      int failedReads = 0;

      while(runflag) 
	{
	  if(ReadChannels()) 
	    {
	      bReadOnce = true;
	      failedReads = 0;

	      if(bPrintData)
		{
		  PrintChannels();
		}
	      
	      if(bListeners) SendSignals();
	    }
	  else
	    {
	      //We didn't have a successful read, but don't do anything
	      //unless we get a bunch in a row.
	      if(++failedReads < kFailedReadsBeforeEcho)
		{
		  continue;
		}
	      //Haven't heard from the Sak2 in a while.  Echo the board to
	      //see if it is still connected and functional
	      else if(sak2v2_echo(10))
		{
		  if(SAK_DEBUG) printf("Echo successful.\n");

		  failedReads = 0;
		  
		  //Board is responding so ask it to verify its configuration.
		  if(!sak2v2_verify_device_configurations())
		    {
		      //Configuration is wrong.  Maybe it was plugged into
		      //a different host.  Try to reconfigure it.
		      sak2v2_configure_devices();
		      
		      if(SAK_DEBUG) printf("Sak2 was incorrectly configured.\n");

		      //Check the reconfiguration.
		      if(!sak2v2_verify_device_configurations())
			{
			  fprintf(stderr, "Sak2 won't configure correctly.  Going through reset procedure.\n");
			  //Something is wrong.  Try startup procedure.
			  break;
			}
		      
		      if(!sak2v2_start_streaming())
			{
			  fprintf(stderr, "Sak2 won't start streaming.  Going through reset procedure.\n");
			  //Something is wrong.  Try startup procedure.
			  break;
			}
		    }
		}
	      else
		{
		  fprintf(stderr, "Sak2 won't respond.  Going through reset procedure.\n");
		  break;
		}
	    }
	  
	  if(bReadOnce) UpdateSignalConnections(&bListeners);
	  
	  killflag = 0;
	}
      
      CloseDownSak2();
    }
  
  // Exiting will close signals.
  
  printf("Exiting nicely.\n");
  return 0;
}

void CloseDownSak2()
{
  int ret;

  if(SAK_DEBUG) printf("CloseDownSak2...");
  
  ret = sak2v2_stop_streaming();
  
  if(SAK_DEBUG) printf("stop_streaming (%d)...", ret);

  ret = sak2v2_clear_configuration();
  
  if(SAK_DEBUG) printf("clear_config (%d)...", ret);
  
  fflush(NULL);
  
  ret = sak2v2_close_serial();
  
  if(SAK_DEBUG) printf("close_serial (%d)...DONE\n", ret);
  
} 
   
int ConfigureSak2Devices()
{
  int ret1, ret2, ret3;

  if(SAK_DEBUG) printf("ConfigureSak2Devices...");

  ret1 = sak2v2_clear_configuration();

  if(SAK_DEBUG) printf("clear_config (%d)...", ret1);

  for(int i = 0; i < SAK2V2_NUM_DEVICES; i++)
    {
      ret2 = sak2v2_set_device((enum sak2v2_devName) i, channelRecs[i].valid);
      
      if(SAK_DEBUG) printf("deviced %d (%d)...", i, ret2);
    }
  
  ret3 = sak2v2_configure_devices();
  
  if(SAK_DEBUG) printf("config_devices (%d)...%s", ret3,(ret1 && ret3) ? "SUCCESS" : "FAILED" );

  return (ret1 && ret3);
}

void PrintUsage()
{
  printf("Usage:Sak2v2Signal"
	 "\n\t-c <a|A|p|e|g|t|i>\t\t#accel0, accel1, polar, ekg, gsr, temp, IR tag"
	 "\n\t-s <context server name>"
	 "\n\t[-u <user id>]"
	 "\n\t[-p]\t\t\t\t#Print data to stdout\n");
  printf("Default options:"
	 "\n\t-s <localhost>\n");
}

void ProcessCommandLine(char *argv[], int argc, char **server, char **user, bool *printData)
{
  bool serverSet(false), userSet(false), channelsAreSet(false);
  char *channels = NULL;
  
  *printData = false;

  for(int c = 0; c < SAK2V2_NUM_DEVICES; c++)
    {
      channelRecs[c].valid = false;
      channelRecs[c].name = NULL;      
    }

  for(int i = 1; i < argc; i++)
    {
      if(argv[i][0] == '-')
	switch(argv[i][1]) {
	case 'p':
	  *printData = true;
	  break;
	case 's':
	  i++;
	  *server = argv[i];
	  printf("Context server is %s.\n", argv[i]);
	  serverSet = true;
	  break;
	case 'u':
	  i++;
	  *user = argv[i];
	  printf("Sak2v2 signal user id %s.\n", argv[i]);
	  userSet = true;
	  break;
	case 'c':
	  i++;
	  channels = argv[i];
	  
	  int k;
	  for(k=0; k < (int) strlen(channels);k++)
	    {
	      switch(channels[k]) {
	      case 'a':
		channelsAreSet = true;
		channelRecs[SAK2V2_ACCEL_0].name = "accel0";
		channelRecs[SAK2V2_ACCEL_0].valid = true;
		channelRecs[SAK2V2_ACCEL_0].reqSampRate = -1;
		printf("Reading 1st accelerometer.\n");
		break;
	      case 'A':
		channelsAreSet = true;
		channelRecs[SAK2V2_ACCEL_1].name = "accel1";
		channelRecs[SAK2V2_ACCEL_1].valid = true;
		channelRecs[SAK2V2_ACCEL_1].reqSampRate = -1;
		printf("Reading 2nd accelerometer.\n");
		break;
	      case 'p':
		channelsAreSet = true;
		channelRecs[SAK2V2_POLAR].name = "polar";
		channelRecs[SAK2V2_POLAR].valid = true;
		channelRecs[SAK2V2_POLAR].reqSampRate = -1;
		printf("Reading Polar heart rate.\n");
		break;
	      case 'i':
		channelsAreSet = true;
		channelRecs[SAK2V2_TAG_READER].name = "tag_reader";
		channelRecs[SAK2V2_TAG_READER].valid = true;
		channelRecs[SAK2V2_TAG_READER].reqSampRate = -1;
		printf("Reading tag reader.\n");
		break;
	      case 'e':
		channelRecs[SAK2V2_EKG].valid = true;
		channelRecs[SAK2V2_EKG].name = "ekg";
		channelRecs[SAK2V2_EKG].reqSampRate = -1;
		channelsAreSet = true; 
		printf("Reading ekg.\n");
		break;
	      case 'g':
		channelRecs[SAK2V2_GSR].valid = true;
		channelRecs[SAK2V2_GSR].name = "gsr";
		channelRecs[SAK2V2_GSR].reqSampRate = -1;
		channelsAreSet = true; 
		printf("Reading gsr.\n");
		break;
	      case 't':
		channelRecs[SAK2V2_TEMP].valid = true;
		channelRecs[SAK2V2_TEMP].name = "temp";
		channelRecs[SAK2V2_TEMP].reqSampRate = -1;
		channelsAreSet = true; 
		printf("Reading temperature.\n");
		break;
	      default:
		fprintf(stderr,"Invalid channel character %c\n",
			  channels[k]);
	      }
	    }
	  break;
	case 'h':
	default:
	  PrintUsage();	      
	  exit(1);
	} 
    }
  
  if(!channelsAreSet) 
    {
      printf("There are no channels set for reading.\n");
      PrintUsage();
      exit(1);
    }

  if(!serverSet) 
    {
      char *eserver = getenv("SAK2V2ZAURUSSIGNAL_SERVER");
      
      if(eserver) *server = eserver;
      else *server = "localhost";
    }
}

void SetupSignalHandlers()
{
 
  struct itimerval *value =
    (struct itimerval *)(malloc(sizeof(struct itimerval)));;

  long loopTime = 20905l - 10000l; // something like 10ms overhead
  //  long loopTime = 31250l - 10000l; // something like 10ms overhead?

  //signal(SIGALRM,SigHandler);
  signal(SIGINT,SigHandler);
  signal(SIGUSR1,SigHandler);
  
  // Set timer value and reset value
  value->it_interval.tv_sec=0;
  value->it_interval.tv_usec=loopTime;
  value->it_value.tv_sec=0;
  value->it_value.tv_usec=loopTime;
  
  // Set real-time timer
  //setitimer(ITIMER_REAL,value,NULL);
  
  // Until we are interrupted...
}


void SigHandler(int sig) {
  switch (sig) {
  case SIGALRM:
    //printf("Signal handler: Alarm received\n");
    break;
  case SIGINT:
    //puts("Signal handler: SIGINT (keyboard interrupt)");
    runflag=0;
    if(killflag) 
      {
	printf("Hard kill, restart Sak2v2 board.\n");
	exit(0);
      }
    else 
      {
	killflag = 1;
      }
    break;
  case SIGUSR1:
    //close nicely and exit
    sak2v2_stop_streaming();
    sak2v2_close_serial();
    
    printf("Killed nicely.\n");
    exit(0);
    break;    
  default:
    printf("Signal handler received %d\n",sig);
  }
  return;
}

void SetupEnchantSignals(char *server, char *user)
{
  char path[128];
  
  for(int i = 0; i < SAK2V2_NUM_DEVICES; i++)
    {
      if(channelRecs[i].valid)
	{
	  
	  if(channelRecs[i].name)
	    {
	      if(user)
		{
		  sprintf(path, "signals/sak2/%s/%s", channelRecs[i].name, user);
		}
	      else
		{
		  sprintf(path, "user/signals/sak2/%s", channelRecs[i].name);
		}
	    }
	  
	  
	  const char *chanUNL = CreateUNL(server, path);
	  
	  channelRecs[i].sig = NewSignalGenUNL(chanUNL);
	  channelRecs[i].feature = NULL;

	  if(channelRecs[i].sig == NULL)
	    {
	      printf("Error creating signal generator for channel %d.\n", i);
	      exit(1);
	    }
	}
    }
}

void UpdateSignalConnections(bool *bListeners)
{
  *bListeners = false;

  // First, accept listeners on all the signals 
  for(int c = 0; c < SAK2V2_NUM_DEVICES; c++)
    {
      if(channelRecs[c].valid && channelRecs[c].feature) 
	{
	  SignalGenAccept(channelRecs[c].sig);
	  
	  if(SignalClientCount(channelRecs[c].sig) > 0)
	    {
	      *bListeners = true;
	    }    
	}
    }
}


int ReadChannels()
{
  if(!sak2v2_read_devices()) return 0;
  
  for(int i = 0; i < SAK2V2_NUM_DEVICES; i++)
    {
      char *bufPtr;
      int bps, num_samp;
      struct timeval readTime;
      
      if(channelRecs[i].valid)
	{
	  if(sak2v2_get_samples((enum sak2v2_devName) i, &bufPtr, &bps, &num_samp, &readTime))
	    {
	      //check to see if this channel has new samples
	      channelRecs[i].freshSample = (num_samp > 0);
	      channelRecs[i].sampTime = readTime;

	      if(channelRecs[i].freshSample)
		{
		  //create a new feature vector if first time
		  if(channelRecs[i].feature == NULL)
		    {
		      //all data read is in the form of characters and to be treated
		      //as ints
		      
		      if(i == SAK2V2_POLAR)
			{
			  //Two samples combined into one
			  SignalSetType(channelRecs[i].sig, T_INT16);
			  SignalSetDimension(channelRecs[i].sig, 1, 1, "bpm", NULL);
			}
		      else
			{
			  SignalSetType(channelRecs[i].sig, T_INT8);
			  SignalSetDimension(channelRecs[i].sig, 1, bps, NULL, NULL);
			}
		      
		      channelRecs[i].feature = (char*) malloc(SignalBufferSize(channelRecs[i].sig));
		    }
		  
		  //Set the new sample
		  if(i == SAK2V2_POLAR)
		    {
		      //int bpm = (5000000/(192*2*16))/(polar0*256 + polar1*60)
		      float bpm = 48828/((float)bufPtr[0]*256 + (float)bufPtr[1]);
		      channelRecs[i].feature[0] = (char) bpm;
		    }
		  else
		    {
		      for(int b = 0; b < bps; b++) 
			{
			  if(i == SAK2V2_ACCEL_0 || i == SAK2V2_ACCEL_1)
			    {
			      channelRecs[i].feature[b] = (signed char)bufPtr[b];
			    }
			  else
			    {
			      channelRecs[i].feature[b] = (unsigned char)bufPtr[b];
			    }
			}
		    }
		}
	    }
	  else
	    {
	      fprintf(stderr, "get_samples for %d channel failed.\n", i);
	      return 0;
	    }
	}
    }
  return 1;
}


void SendSignals()
{
  for(int i = 0; i < SAK2V2_NUM_DEVICES; i++)
    {
      if(channelRecs[i].valid && 
	 (SignalClientCount(channelRecs[i].sig) > 0) &&
	 channelRecs[i].feature &&
	 channelRecs[i].freshSample)
	{
	  
	  SignalSend(channelRecs[i].sig, &channelRecs[i].sampTime, channelRecs[i].feature);
	}
    }
}


void PrintChannels()
{
  for(int i = 0; i < SAK2V2_NUM_DEVICES; i++)
    {
      if(channelRecs[i].valid &&
	 channelRecs[i].feature &&
	 channelRecs[i].freshSample)
	{
	  if(channelRecs[i].name) printf("%s: ", channelRecs[i].name);
	  else printf("dev %d: ", i);
	  
	  PrintSignalElements(stdout, channelRecs[i].sig, 
			      channelRecs[i].sampTime, channelRecs[i].feature);
	  

	  /*

	  printf("[ ");
	  for(int e = 0; e < channelRecs[i].feature->getNumElements(); e++)
	    {
	      printf("%ld ", channelRecs[i].feature->getElement(e));
	    }
	  
	  printf("]\n");
	  */
	}
    }
}













