/*****************************************************************************
 * File: AccelFeature.c
 * Author: Mark Blum (mark@media.mit.edu), MIT Wearable Computing Group
 * Birth Date: 06/05/05
 * http://www.media.mit.edu/wearables/
 * 
 * Copyright (C) 2005 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************
 * This program computes features from acceleration signals. Originally,
 * only an FFT feature was computed. The mean and variance features were added 
 * later. The calculation is hard-coded for 1 to 2 accelerometers. Becasue of
 * different sampling rates, thoughtful programming is required. This version
 * succedes without multi-threading. If more sensors are introduced, multi-
 * threading will probably be necessary.
 * How it works: The values are continuously stored in ring-buffers. Every
 * TIMESTEP seconds, the last WINDOWSIZE samples are used to calculate mean
 * and variance, which are then posted again as a signal. In the case of two
 * sensors, the program will read values of Acc_1 in a loop until the timestamps
 * excede the next TIMESTEP. Then, in a second loop, the values of Acc_2 are read
 * until those timestamps reach the same time. This is only possible due to the 
 * buffering functionality of Enchantment signals. If any input signal goes down
 * the program will stall.
 * The mean and variance features were verified as exact compared to Matlab
 * The Signals are expected to be posted by MITeSignal
 * NOTE: Calibration is vital with the use of different accelerometers!
 * The Calibration file which is loaded here was computed by MITeCalibrate
 *****************************************************************************
 * Description of the FFT feature: 
 *
 * Every 32 samples a 64-element box-function window fourier transform
 * is computed (50% overlap per window).  This results in a 64 element
 * DFFT vector. Note that the window size (and, therefore, the
 * resulting frequency range and lag time) is just a constant here,
 * and not actually determined from the data.
 * From the DFFT vector, the power spectrum is computed. The power
 * spectrum is the magnitude square of the fourier transform. The
 * fourier transform contains phase information and is a reversable
 * transform.  The power spectrum throws away phase information (and
 * has half the terms) of the fourier transform, and is not
 * reversible.
 * The result is a signal of 31-element power spectrum vectors (power
 * spectrum minus the DC component).
 *****************************************************************************/
 
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <math.h>

#ifdef FFTW3
#include <fftw3.h>
#else
#include <rfftw.h>
#endif

#include <enchantment/signals.h>

#define INPUT_DIM 256
#define RING_BUFF_SIZE 776 //2*WINDOWSIZE + 8
#define WINDOWSIZE 384 //3*128, same as in Matlab
#define TIMESTEP 2

#define FFT32 0
// Computes the FFT over the magnitude of the 3-axes in a window of 64 samples --> 31 Element Vector
#define FFT64 1

#define ACC1_MEAN 2
// Computes the mean for each of the 3 axes of the hip-worn sensor --> 3 Element Vector
#define ACC1_MEAN_VAR 3
// Computes the mean and variance for each of the 3 axes of the hip worn sensor --> 6 Element Vector
#define ACC2_MEAN 4
// Computes the mean for each of the 3 axes of the wrist-worn sensor --> 3 Element Vector
#define ACC2_MEAN_VAR 5
// Computes the mean and variance for each of the 3 axes of the wrist-worn sensor --> 6 Element Vector
#define ACC12_MEAN 6
// Computes the mean for each of the 3 axes of both sensors 
//  --> 6 Element Vector [meanX1 meanY1 meanZ1 meanX2 meanY2 meanZ2]
#define ACC12_MEAN_VAR 7
// Computes the mean and variance for each of the 3 axes of both sensors 
//  --> 12 Element Vector [meanX1 meanY1 meanZ1 meanX2 meanY2 meanZ2 varX1 varY1 varZ1 varX2 varY2 varZ2]

int outFeatureSizes[] = {32, 64, 3, 6, 3, 6, 6, 12};
// #define MEAN_X1 263
// #define MEAN_Y1 274 
// #define MEAN_Z1 268
// #define MEAN_X2 232
// #define MEAN_Y2 234
// // //#define MEAN_Y2 253
// #define MEAN_Z2 288
// #define SCALE_1G 20
// #define SCALE_SQ 400
// // #define MEAN_X2 228
// // #define MEAN_Y2 226
// // #define MEAN_Z2 288


      
inline
float Magnitude(short *inVec) {
    float x= (float)inVec[0];
    float y= (float)inVec[1];
    float z= (float)inVec[2];
  //cout << *feature << endl;
    return sqrt(x*x + y*y + z*z);
}

inline
float CalcMean(short *inVec, int windowSize) {
    int i,sum = 0;
    for (i=0; i < windowSize; i++){
        sum += inVec[i];
    }
    return (float)sum/((float)windowSize);
}

inline
float CalcVar(short *inVec, int windowSize, float mean) {
    int i;
    float sum = 0;
    for (i=0; i < windowSize; i++){
        sum += inVec[i]*inVec[i];
    }
    return sum/((float)windowSize) - mean*mean;            // var(X) = E(X - u)^2 = E(X^2) - u^2
}

inline
float CalcMean2(short *inVec, int index, int windowSize) {
    int i,sum = 0;
    for (i=0; i < windowSize; i++){
        index = (index > 0) ? index -1 : RING_BUFF_SIZE-1;
        sum += inVec[index];
    }
    return (float)sum/((float)windowSize);
}

inline
float CalcVar2(short *inVec, int index, int windowSize, float mean) {
    int i;
    float sum = 0;
    for (i=0; i < windowSize; i++){
        index = (index > 0) ? index -1 : RING_BUFF_SIZE-1;
        sum += inVec[index]*inVec[index];
    }
    return sum/((float)windowSize) - mean*mean;            // var(X) = E(X - u)^2 = E(X^2) - u^2
}

#ifdef FFTW3
typedef double fftw_real;
#endif

void PrintUsage()
{
    printf("Usage: AccelFeature -i [unl1] [unl2] -o [unl] -m [mode] -c [calibration file] -g [0/1]\n");
    printf("                                                                   \n");
    printf("       -i Input UNLs of accelermeter signals \n");
    printf("       -o Output UNL of feature vector signal \n");
    printf("       -m Mode\n");
    printf("       -c File with calibration values, created by MITeCalibrate\n");
    printf("       -g Shall Y1 be averaged to [0 0 0] i.e. the means of Acc1 would all be \n");
    printf("          about 0 if Acc1 is worn on the hip. Required by some classifiers.   \n");
    printf("                             \n");
    printf("       Defaults:  -i [/signals/MITE/MITe_5] [/signals/MITE/MITe_9] \n");
    printf("                  -o [/features/accel]\n");
    printf("                  -m [acc12_mean_var]\n");
    printf("                  -c [Config_MITes.txt]\n");
    printf("                  -g [0]\n");
    printf("\n");
}

int main(int argc, char *argv[]) 
{
    FILE* fid;
    char *inUNL1 = "/signals/MITe/MITe_5";
    char *inUNL2 = "/signals/MITe/MITe_9";
    char *outUNL = "/features/accel";
    char *calibrationFile = "Config_MITes.txt";
    int mode = ACC12_MEAN_VAR;
    int twoSensors = 1;
    int acc1_1G_Correction = 0;

    short inBuffer[3];
    
    int oldindex1, oldindex2;
    
    float *outBuffer = NULL;
    
    Signal *inSignal1;
    Signal *inSignal2;
    Signal *outSignal;

    short valsX1[RING_BUFF_SIZE], valsY1[RING_BUFF_SIZE], valsZ1[RING_BUFF_SIZE];
    float meanX1, meanY1, meanZ1, varX1, varY1, varZ1;
    
    short valsX2[RING_BUFF_SIZE], valsY2[RING_BUFF_SIZE], valsZ2[RING_BUFF_SIZE];
    float meanX2, meanY2, meanZ2, varX2, varY2, varZ2;
    
    /*                      | means      | 1G        | 1G ^ 2             */
    float calibration1[] = {237, 284, 235, 60, 57, 60, 60*60, 57*57, 60*60};
    float calibration2[] = {299, 311, 275, 60, 60, 56, 60*60, 60*60, 56*56};
       
    int index1, index2, i, first = 1;
    long count;
    //float t_diff;
  
    double vals[INPUT_DIM];
    fftw_real invals[INPUT_DIM];
    fftw_real outvals[INPUT_DIM];
    double prescalar = 1;

#ifdef FFTW3
        fftw_plan plan;
#else
        rfftw_plan plan;
#endif
        
    struct timeval inTimestamp, inTimestamp2, outTimestamp;
  
    ///// Parse input arguments ///////////////////////////////////////////

    i = 1;
    while ( i < argc )
    {
        if (argv[i][0] == '-' )
        {
            i++;
            if ( i == argc )
            {
                PrintUsage();
                return -1;
            }
            switch ( argv[i-1][1] )
            {
                case 'i': {
                    inUNL1 = argv[i];
                    twoSensors = 0;
                    if ( i < argc-1 ){
                        if (argv[i+1][0] != '-' ){
                            inUNL2 = argv[++i];
                            twoSensors = 1;
                        }
                    }
                    break;
                }
                case 'o': {
                    outUNL = argv[i];
                    break;
                }
                case 'm': {
                    mode = -1;
                    if(strcmp(argv[i],"acc1_mean")==0){
                        mode = ACC1_MEAN;
                    }
                    if(strcmp(argv[i],"acc1_mean_var")==0){
                        mode = ACC1_MEAN_VAR;
                    }
                    if(strcmp(argv[i],"acc2_mean")==0){
                        mode = ACC2_MEAN;
                    }
                    if(strcmp(argv[i],"acc2_mean_var")==0){
                        mode = ACC2_MEAN_VAR;
                    }
                    if(strcmp(argv[i],"acc12_mean_var")==0){
                        mode = ACC12_MEAN_VAR;
                    }
                    if(strcmp(argv[3],"fft64")==0){
                        mode = FFT64;
#ifdef FFTW3
        plan = fftw_plan_r2r_1d(INPUT_DIM, invals, outvals, FFTW_R2HC, FFTW_PATIENT);
#else
        plan = rfftw_create_plan(INPUT_DIM, FFTW_REAL_TO_COMPLEX, FFTW_MEASURE);
#endif
                    }
                    if (mode == -1) {
                        printf("Unknown mode '%s'\n",argv[i]); 
                        PrintUsage();
                        exit(1);
                    }
                    
                    break;
                }
                case 'c': {
                    calibrationFile = argv[i];
                    break;
                }
                case 'g': {
                    acc1_1G_Correction = atoi(argv[i]);
                    break;
                }
                default: PrintUsage(); return -1;
            }
            i++;
        }
        else 
        { 
            PrintUsage();
            return -1;
        }
    }  ///// end parsing input arguments
    
    printf("AccelFeature: inUNL1: %s  inUNL2: %s  outUNL: %s  mode: %d  calib: %s \n", inUNL1, (twoSensors) ? inUNL2 : "none",outUNL,mode,calibrationFile);

    fid = fopen(calibrationFile,"r");
    if (fid == NULL) {
        printf("AccelFeature: Warning, could not find file %s\n", calibrationFile);
        printf("              Using defualt calibration values\n");
    }    
    else
    {
        fscanf(fid,"%f %f %f %f %f %f",calibration1, calibration1+1, calibration1+2, calibration1+3, calibration1+4, calibration1+5);
        calibration1[6] = calibration1[3]*calibration1[3]; calibration1[7] = calibration1[4]*calibration1[4]; calibration1[8] = calibration1[5]*calibration1[5];
        if (acc1_1G_Correction) calibration1[1] = calibration1[1]+calibration1[4];
        if (twoSensors){
            fscanf(fid,"%f %f %f %f %f %f",calibration2, calibration2+1, calibration2+2, calibration2+3, calibration2+4, calibration2+5);
            calibration2[6] = calibration2[3]*calibration2[3]; calibration2[7] = calibration2[4]*calibration2[4]; calibration2[8] = calibration2[5]*calibration2[5];
        }
        fclose(fid);   
    }
        
    inSignal1 = NewSignalUNL(inUNL1);
    if(inSignal1 == NULL) {
        printf("AccelFeature: Error allocating input signal %s\n",inUNL1);
        exit(1);
    }
          
    outSignal = NewSignalGenUNL(outUNL);
    if(outSignal == NULL) {
        printf("AccelFeature: Error allocating outSignal %s\n",outUNL);
        exit(1);
    }
    SignalSetType(outSignal, T_FLOAT);
    SignalSetDimension(outSignal, 1, outFeatureSizes[mode], NULL, NULL);
    //SignalSetSampleRate(out, 0, 1);        //old
    outBuffer = (float*) malloc(SignalBufferSize(outSignal));
    
    for(i=0; i<RING_BUFF_SIZE; i++) {
        valsX1[i] = 0;
        valsY1[i] = 0;
        valsZ1[i] = 0;
        valsX2[i] = 0;
        valsY2[i] = 0;
        valsZ2[i] = 0;
    }
    for(i=0; i<INPUT_DIM; i++) {
        vals[i] = 0;
    }
    if (!SignalConnectWait(inSignal1)){
        printf("AccelFeature: Error connecting to %s\n",inUNL1);
        exit(1);
    }
    
    if (SignalGetType(inSignal1) != T_INT16 || SignalGetLength(inSignal1, 1) != 3)
    {
        printf("in signal type error: type: #%d; columns: %d\n",
               SignalGetType(inSignal1), SignalGetLength(inSignal1, 1));
        exit(1);
    }

   // SignalSetType(in, 4, SIGNAL_TYPE_INT);
//     fid = fopen("./data/features.txt","w");
//     if (fid == NULL) {
//         printf("AccelFeature: Error, could not create file ./data/features.txt\n");
//         exit(1);
//     }

    printf("debug2\n");
    
    if (twoSensors)
    {
        inSignal2 = NewSignalUNL(inUNL2);
        if(inSignal2 == NULL) {
            printf("AccelFeature: Error allocating inputsignal%s\n",inUNL2);
            exit(1);
        }
        if (!SignalConnectWait(inSignal2)){
            printf("AccelFeature: Error connecting to %s\n",inUNL2);
            exit(1);
        }
        
        printf("AccelFeature: Starting feature calculation for 2 Signals\n");
        
        index1 = 0;
        index2 = 0;
        count = 0;
        gettimeofday(&outTimestamp, NULL);
        while (1) 
        {
            SignalGenAccept(outSignal);
            oldindex1 = index1;
            oldindex2 = index2;
            do {
                SignalRecv(inSignal1, &inTimestamp, (void *) inBuffer);
                                
                valsX1[index1] = inBuffer[0];
                valsY1[index1] = inBuffer[1];
                valsZ1[index1] = inBuffer[2];
                
                index1 = (index1 + 1) % RING_BUFF_SIZE;
                
                if ((inTimestamp.tv_sec > outTimestamp.tv_sec) || 
                     (inTimestamp.tv_sec == outTimestamp.tv_sec && inTimestamp.tv_usec > outTimestamp.tv_usec) )
                {
                    //printf("--Read Acc1, index1 = %d\n",index1);
                    break;
                }
            }
            while(1);
            do {
                SignalRecv(inSignal2, &inTimestamp2, (void *) inBuffer);
                                
                valsX2[index2] = inBuffer[0];
                valsY2[index2] = inBuffer[1];
                valsZ2[index2] = inBuffer[2];
                
                index2 = (index2 + 1) % RING_BUFF_SIZE;
                
                if ( (inTimestamp2.tv_sec > outTimestamp.tv_sec) || 
                      (inTimestamp2.tv_sec == outTimestamp.tv_sec && inTimestamp2.tv_usec > outTimestamp.tv_usec) )
                {
                    //printf("--Read Acc2, index2 = %d\n",index2);
                    break;
                }
            }
            while(1);
        
            //printf("Samples used for this slice: delta 1 = %d     delta 2 = %d\n",((index1-oldindex1)+RING_BUFF_SIZE)%RING_BUFF_SIZE,((index2-oldindex2)+RING_BUFF_SIZE)%RING_BUFF_SIZE);
            
            ///// start action when we reach half a window ///////////////////////////////
        
            if ((mode == ACC12_MEAN) || (mode == ACC12_MEAN_VAR)) {
                meanX1 = CalcMean2(valsX1, index1, WINDOWSIZE);
                meanY1 = CalcMean2(valsY1, index1, WINDOWSIZE);
                meanZ1 = CalcMean2(valsZ1, index1, WINDOWSIZE);
                meanX2 = CalcMean2(valsX2, index2, WINDOWSIZE);
                meanY2 = CalcMean2(valsY2, index2, WINDOWSIZE);
                meanZ2 = CalcMean2(valsZ2, index2, WINDOWSIZE);
                outBuffer[0] = (meanX1-calibration1[0])/calibration1[3];
                outBuffer[1] = (meanY1-calibration1[1])/calibration1[4];
                outBuffer[2] = (meanZ1-calibration1[2])/calibration1[5];
                outBuffer[3] = (meanX2-calibration2[0])/calibration2[3];
                outBuffer[4] = (meanY2-calibration2[1])/calibration2[4];
                outBuffer[5] = (meanZ2-calibration2[2])/calibration2[5];
                if (mode == ACC12_MEAN_VAR) {
                    varX1 = CalcVar2(valsX1, index1, WINDOWSIZE, meanX1);
                    varY1 = CalcVar2(valsY1, index1, WINDOWSIZE, meanY1);
                    varZ1 = CalcVar2(valsZ1, index1, WINDOWSIZE, meanZ1);
                    varX2 = CalcVar2(valsX2, index2, WINDOWSIZE, meanX2);
                    varY2 = CalcVar2(valsY2, index2, WINDOWSIZE, meanY2);
                    varZ2 = CalcVar2(valsZ2, index2, WINDOWSIZE, meanZ2);
                    outBuffer[6] = varX1/calibration1[6];
                    outBuffer[7] = varY1/calibration1[7];
                    outBuffer[8] = varZ1/calibration1[8];
                    outBuffer[9] = varX2/calibration2[6];
                    outBuffer[10] = varY2/calibration2[7];
                    outBuffer[11] = varZ2/calibration2[8];
                }
            }
            //t_diff = (float)(inTimestamp2.tv_sec - inTimestamp.tv_sec) + ((float)(inTimestamp2.tv_usec - inTimestamp.tv_usec))/1e6;
            //fprintf(stderr, "t2 - t1 = %6.6f\n",t_diff);
            
            //printf("%f %f %f %f %f %f %f %f %f %f %f %f\n",outBuffer[0],outBuffer[1],outBuffer[2],outBuffer[3],outBuffer[4],outBuffer[5],outBuffer[6],outBuffer[7],outBuffer[8],outBuffer[9],outBuffer[10],outBuffer[11]);
            
            
//            if((count % 100) == 0){
//                fclose(fid);    
//                fid = fopen("./data/features.txt","a");
//            }
//            fprintf(fid,"%06ld.%06ld", outTimestamp.tv_sec, outTimestamp.tv_usec);  
//            fprintf(fid," %g %g %g %f %f %f %f %f %f %f %f %f\n",outBuffer[0],outBuffer[1],outBuffer[2],outBuffer[3],outBuffer[4],outBuffer[5],outBuffer[6],outBuffer[7],outBuffer[8],outBuffer[9],outBuffer[10],outBuffer[11]);
                        
            SignalGenAccept(outSignal);
            SignalSend(outSignal, &outTimestamp, (char *)outBuffer);
            //outTimestamp = inTimestamp;
            
                       
            outTimestamp.tv_sec = outTimestamp.tv_sec + TIMESTEP;
          
            //if((count % 1000) == 0) printf("AccelFeature: %ld entries\n", count);
            count++;
        }
        fclose(fid);
        
    }
    else   //(!twoSensors)
    {
    
        printf("AccelFeature: Starting feature calculation for 1 Signal\n");
        index1 = 0;
        count = 0;
        while (1) 
        {
            SignalGenAccept(outSignal);
            SignalRecv(inSignal1, &inTimestamp, (void *) inBuffer);
            
            ///// Store Values in a ring-buffer ///////////////////////////////////////////

            if (mode == FFT64){      
                vals[index1] = Magnitude(inBuffer);
            }
            else {
                valsX1[index1] = inBuffer[0];
                valsY1[index1] = inBuffer[1];
                valsZ1[index1] = inBuffer[2];
            }
            index1 = (index1 + 1) % INPUT_DIM;
        
            ///// start action when we reach half a window ///////////////////////////////
        
            if (index1 == 0 || (index1 == INPUT_DIM/2 && !first))
            {
                first = 0;
            
                if (mode == FFT64) {
                    for (i = 0; i < index1; i++)
                    {
                        invals[i] = vals[i + INPUT_DIM - index1];
                    }
                    for (i = index1; i < INPUT_DIM; i++)
                    {
                        invals[i] = vals[i - index1];
                    }
#ifdef FFTW3
                fftw_execute(plan);
#else
                rfftw_one(plan, invals, outvals);
#endif
                for (i = 1; i < (INPUT_DIM+1)/2; i++)    // skip DC comp, (i < INPUT_DIM/2 rounded up) 
{
                    // magnitude squared, by squaring the real and imaginary parts and adding them.
    outBuffer[i - 1] = (outvals[i] * outvals[i] + outvals[INPUT_DIM-i] * outvals[INPUT_DIM-i]) * prescalar;
}
                outBuffer[INPUT_DIM/2] = outvals[INPUT_DIM/2] * outvals[INPUT_DIM/2] * prescalar;
                }

            
            // we need not consider the correct sequence of values, since it does not influence means and variances          
                if ((mode == ACC1_MEAN) || (mode == ACC1_MEAN_VAR)) {
                    meanX1 = CalcMean(valsX1, INPUT_DIM);
                    meanY1 = CalcMean(valsY1, INPUT_DIM);
                    meanZ1 = CalcMean(valsZ1, INPUT_DIM);
                    outBuffer[0] = (meanX1-calibration1[0])/calibration1[3];
                    outBuffer[1] = (meanY1-calibration1[1])/calibration1[4];
                    outBuffer[2] = (meanZ1-calibration1[2])/calibration1[5];
                    if (mode == ACC1_MEAN_VAR) {
                        varX1 = CalcVar(valsX1, INPUT_DIM, meanX1);
                        varY1 = CalcVar(valsY1, INPUT_DIM, meanY1);
                        varZ1 = CalcVar(valsZ1, INPUT_DIM, meanZ1);
                        outBuffer[3] = varX1/calibration1[6];
                        outBuffer[4] = varY1/calibration1[7];
                        outBuffer[5] = varZ1/calibration1[8];
                    }
                }
                if ((mode == ACC2_MEAN) || (mode == ACC2_MEAN_VAR)) {
                    meanX2 = CalcMean(valsX2, INPUT_DIM);
                    meanY2 = CalcMean(valsY2, INPUT_DIM);
                    meanZ2 = CalcMean(valsZ2, INPUT_DIM);
                    outBuffer[0] = (meanX2-calibration2[0])/calibration2[3];
                    outBuffer[1] = (meanY2-calibration2[1])/calibration2[4];
                    outBuffer[2] = (meanZ2-calibration2[2])/calibration2[5];
                    if (mode == ACC2_MEAN_VAR) {
                        varX2 = CalcVar(valsX2, INPUT_DIM, meanX2);
                        varY2 = CalcVar(valsY2, INPUT_DIM, meanY2);
                        varZ2 = CalcVar(valsZ2, INPUT_DIM, meanZ2);
                        outBuffer[3] = varX2/calibration2[6];
                        outBuffer[4] = varY2/calibration2[7];
                        outBuffer[5] = varZ2/calibration2[8];
                    }
                }
          
                SignalGenAccept(outSignal);
                SignalSend(outSignal, &inTimestamp, (char *)outBuffer);
                //outTimestamp = inTimestamp;
          
                if((count % 2000) == 0) printf("AccelFeature: %ld entries\n", count);
                count++;
            }
            else if (index1 == INPUT_DIM/2)
            {
            // first timestamp
            //outTimestamp = inTimestamp;
            }
        }
    }
    return 0;
}

