/*****************************************************************************
 * File: C45Classifier.c
 * Author: Mark Blum (mark@media.mit.edu), MIT Wearable Computing Group
 * Birth Date: 07/09/05
 * http://www.media.mit.edu/wearables/
 * 
 * Copyright (C) 2005 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************
 * This is the Enchantment interface to the external c4.5 code. It takes care
 * of reading feature signals, consuting the c4.5 functions, and posting the 
 * results as signals or nodes on the whiteboard. File I/O is also supported
 * for offline use.
 * The code looks like a mess. This is because the c45 packages used has strange
 * coding conventions and could not be easily integrated.
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

#include <enchantment/signals.h>
#include <enchantment/node.h>
#include <enchantment/whiteboard.h>


/////////////////////////////////////////////////////////////
/// c45-code
/// The reason for all this being here is runtime problems
/// with the nicely seperated version.
/////////////////////////////////////////////////////////////

#include "c45_extern.i"

/////  External data  -- see c4.5.c for meanings ///
short MaxAtt, MaxClass, MaxDiscrVal;
ItemNo MaxItem;
Description* Item;
DiscrValue* MaxAttVal;
String *ClassName, *AttName, **AttValName, FileName;
char* SpecialStatus;
typedef struct ValRange *RangeDescRec;
struct ValRange
{
    Boolean Known,    /* is range known? */
    Asked;            /* has it been asked? */
    float LowerBound, /* lower bound given */
    UpperBound,       /* upper ditto */
    *Probability;      /* prior prob of each discr value */
};
RangeDescRec RangeDesc;
Tree   DecisionTree,  /* tree being used */
GetTree();
float *LowClassSum,    /* accumulated lower estimates */
      *ClassSum = Nil; /* accumulated central estimates */
short VERBOSITY;
#define Fuzz 0.01  /* minimum weight */

void GetNames(void);
//void ClassifyCase(Tree Subtree, float Weight);
//void Decision(ClassNo c, float p, float lb, float ub);

    ClassifyCase(Subtree, Weight)
/*  ------------ 	 */
    Tree Subtree;
    float Weight;
{
    DiscrValue v;
    float BranchWeight, Area(), Interpolate();
    Attribute a;
    short s;
    ClassNo c;
    
    /*  A leaf  */

    if ( ! Subtree->NodeType )
    {
	Verbosity(1)
	    printf("\tClass %s weight %g cases %g\n", 
		    ClassName[Subtree->Leaf], Weight, Subtree->Items);

	if ( Subtree->Items > 0 )
	{
	    /*  Adjust class sum of ALL classes, but adjust low class sum
		of leaf class only  */

	    ForEach(c, 0, MaxClass)
	    {
		ClassSum[c] += Weight * Subtree->ClassDist[c] / Subtree->Items;
	    }

	    LowClassSum[Subtree->Leaf] +=
		Weight * (1 - Subtree->Errors / Subtree->Items);
	}
	else
	{
	    ClassSum[Subtree->Leaf] += Weight;
	}

	return;
    }

    a = Subtree->Tested;

    //this need not be done in our version because RangeDesc[] is filled with data from the function calling ClassifyCase()
    //CheckValue(a, Subtree);

    /*  Unknown value  */

    if ( ! RangeDesc[a].Known )
    {
	ForEach(v, 1, Subtree->Forks)
	{
	    ClassifyCase(Subtree->Branch[v],
		     (Weight * Subtree->Branch[v]->Items) / Subtree->Items);
	}
	return;
    }

    /*  Known value  */

    switch ( Subtree->NodeType )
    {
	case BrDiscr:  /* test of discrete attribute */

	    ForEach(v, 1, MaxAttVal[a])
	    {
		BranchWeight = RangeDesc[a].Probability[v];
		if ( BranchWeight > 0 )
		{
		    Verbosity(1)
		    	printf("\tWeight %g: test att %s (val %s = %g)\n",
		    	       Weight, AttName[a], AttValName[a][v],
			       BranchWeight);

		    ClassifyCase(Subtree->Branch[v], Weight * BranchWeight);
		}
	    }
	    break;

	case ThreshContin:  /* test of continuous attribute */

	    BranchWeight = 
		RangeDesc[a].UpperBound <= Subtree->Lower ? 1.0 :
		RangeDesc[a].LowerBound > Subtree->Upper ? 0.0 :
		RangeDesc[a].LowerBound != RangeDesc[a].UpperBound ?
		    (Area(Subtree, RangeDesc[a].LowerBound) -
		     Area(Subtree, RangeDesc[a].UpperBound)) /
		    (RangeDesc[a].UpperBound - RangeDesc[a].LowerBound) :
		Interpolate(Subtree, RangeDesc[a].LowerBound) ;

	    Verbosity(1)
	        printf("\tWeight %g: test att %s (branch weight=%g)\n",
	    	       Weight, AttName[a], BranchWeight);

	    if ( BranchWeight > Fuzz )
	    {
		ClassifyCase(Subtree->Branch[1], Weight * BranchWeight);
	    }
	    if ( BranchWeight < 1-Fuzz )
	    {
		ClassifyCase(Subtree->Branch[2], Weight * (1 - BranchWeight));
	    }
	    break;

	case BrSubset:  /* subset test on discrete attribute  */

	    ForEach(s, 1, Subtree->Forks)
	    {
		BranchWeight = 0.0;
		ForEach(v, 1, MaxAttVal[a])
		{
		    if ( In(v, Subtree->Subset[s]) )
		    {
			BranchWeight += RangeDesc[a].Probability[v];
		    }
		}
		if ( BranchWeight > 0 )
		{
		    Verbosity(1)
		    	printf("\tWeight %g: test att %s (val %s = %g)\n",
		    	       Weight, AttName[a], AttValName[a][v],
			       BranchWeight);

		    ClassifyCase(Subtree->Branch[s], Weight * BranchWeight);
		}
	    }
	    break;
    }
}

float Interpolate(t, v)
/*    ---- 	 */
    Tree t;
    float v;
{
    float Sum=Epsilon;

    if ( v <= t->Lower )
    {
	return 1.0;
    }

    if ( v <= t->Cut )
    {
	return 1 - 0.5 * (v - t->Lower) / (t->Cut - t->Lower + Epsilon);
    }

    if ( v < t->Upper )
    {
	return 0.5 - 0.5 * (v - t->Cut) / (t->Upper - t->Cut + Epsilon);
    }

    return 0.0;
}

float Area(t, v)
/*    ---- 	 */
    Tree t;
    float v;
{
    float Sum=Epsilon, F;

    if ( v < t->Lower )
    {
	Sum += t->Lower - v;
	v = t->Lower;
    }

    if ( v < t->Cut )
    {
	F = (t->Cut - v ) / (t->Cut - t->Lower + Epsilon);

	Sum += 0.5 * (t->Cut - v) + 0.25 * F * (t->Cut - v);
	v = t->Cut;
    }

    if ( v < t->Upper )
    {
	F = (t->Upper - v ) / (t->Upper - t->Cut + Epsilon);

	Sum += 0.25 * (t->Upper - v) * F;
    }

    Verbosity(1) printf("lower=%g  cut=%g  upper=%g  area=%g\n",
    			t->Lower, t->Cut, t->Upper, Sum);

    return Sum;
}

    Decision(c, p, lb, ub)
/*  -------- 	 */
    ClassNo c;
    float p, lb, ub;
{
    printf("\t%s", ClassName[c]);

    if ( p < 1-Fuzz || lb < ub - Fuzz )
    {
	printf("  CF = %.2f", p);
	if ( lb < ub - Fuzz )
	{
	    printf("  [ %.2f - %.2f ]", lb, ub);
	}
    }

    printf("\n");
}

/////////////////////////////////////////////////////////////
/// end c45-code
/////////////////////////////////////////////////////////////

//Static so we can flush it on
FILE* outFile_fid;
volatile int runflag = 1;
volatile int killflag = 0;

void PrintUsage()
{
    printf("Usage: C45Classifier [OPTION]...\n");
    printf("                                                  \n");
    printf("       -m [FILE]     filestem for C45 model\n");
    printf("       -i [UNL]      read input signal from [UNL] \n");
    printf("       -o [UNL]      write output signal to [UNL] \n");
    printf("       -p [UNL]      post output to [UNL] on whiteboard\n");
    printf("       -r [FILE]     read input from [FILE] \n");
    printf("       -w [FILE]     write output to [FILE] \n");
    printf("       -v            verbose, write output to stdout \n");
    printf("       -t            use current system time for output timestamps\n");
    printf("       In case of whiteboard posting: \n");
    printf("       -d            detailed post, includes a node per class with its\n");
    printf("                     probability, a node with probability vector and a \n");
    printf("                     decision node with a string of the most likely class.\n");
    printf("                     Per default, only the vector is posted.\n");
    printf("       In case file I/O: \n");
    printf("       -n            no timestamps in input and output files\n");
    printf("                                            \n");
}

void SigHandler(int sig) {
    switch (sig) {
        case SIGINT:
            runflag=0;
            if(killflag)
            {
                fflush(stdout);
                if (outFile_fid!=NULL){
                    fflush(outFile_fid);
                }
                fprintf(stderr,"Hard kill, flushing output streams and exiting.\n");
                exit(0);
            }
            else
            {
                killflag = 1;
            }
            
            break;
            default:
                printf("Signal handler received %d\n",sig);
    }
    return;
}

int main(int argc, char *argv[]) {
    FILE* inFile_fid;
    char *inUNL = NULL; // (ex. /signals/feat)
    char *outUNL = NULL;   // (ex. /classifiers/posture)
    char *postUNL = NULL;   // (ex. /classifiers/posture)
    char *inFile = NULL;
    char *outFile = NULL;
    char *modelFile = NULL;
    int verbose = 0;
    int useCurrentTime = 0;
    int detailedPost = 0;
    int fileTimeStamps = 1;
    float* inBuffer = NULL;
    double* outBuffer = NULL;
    int numFeatures, numClasses;
    int endOfFile = 0;
    Signal *inSignal = NULL;
    Signal *outSignal = NULL;
    struct timeval time;
    int i,n;
    char temp[128];
    NodeType *updateNode = NULL;
    NodeType *labelNode = NULL;
    PublishNodeInfo *pubInfo = NULL;
    Vector *probabilityVector = NULL;
    char *nodeName;
    
     ///// C4.5 variables /////
    Attribute a;
    ClassNo c, BestClass;
    float Uncertainty=1.0;
    
    signal(SIGINT,SigHandler);  //die nicely
    printf("C45Classifier start\n");
    
    ///// Parse input arguments ///////////////////////////////////////////
    
    i = 1;
    while ( i < argc ) {
        if (argv[i][0] == '-' ) {
            switch ( argv[i][1] ) {  // options without arguments
                case 'v': {
                    verbose = true;
                    break;
                }
                case 't': {
                    useCurrentTime = true;
                    break;
                }
                case 'd': {
                    detailedPost = true;
                    break;
                }
                case 'n': {
                    fileTimeStamps = false;
                    break;
                }
                default: {
                    i++;
                    if ( i == argc ) // check if argument follows
                    {
                        PrintUsage();
                        exit(1);
                    }
                    switch ( argv[i-1][1] ) { // options with arguments
                        case 'i': {
                            inUNL = argv[i];
                            break;
                        }
                        case 'o': {
                            outUNL = argv[i];
                            break;
                        }
                        case 'p': {
                            postUNL = argv[i];
                            break;
                        }
                        case 'r': {
                            inFile = argv[i];
                            break;
                        }
                        case 'w': {
                            outFile = argv[i];
                            break;
                        }
                        case 'm': {
                            modelFile = argv[i];
                            break;
                        }
                        default: {
                            printf("Unknown option -%c\n\n",argv[i-1][1]);
                            PrintUsage();
                            exit(1);
                        }
                    } // end options with arguments
                }
            } //end switch ( argv[i][1]
            i++;
        }
        else  // no '-' detected
        {
            PrintUsage();
            exit(1);
        }
    }  ///// end parsing input arguments
       
    if (inUNL==NULL && inFile==NULL){
        printf("C45Classifier: Error, no input specified!\n\n");
        PrintUsage(); 
        exit(1);
    }
    if (modelFile==NULL){
        printf("C45Classifier: Error, no model file specified!\n\n");
        PrintUsage(); 
        exit(1);
    }
    FileName = modelFile;
    
    if (inUNL!=NULL) {
        inSignal = NewSignalUNL(inUNL);
        if(inSignal == NULL) {
            printf("C45Classifier: Error allocating signal %s\n",inUNL);
            exit(1);
        }
    }
    else
    {
        inFile_fid = fopen(inFile,"r");
        if (inFile_fid == NULL) {
            printf("C45Classifier: Error, could not open file %s\n", inFile);
            exit(1);
        }
    }
    
    printf("C45Classifier: Loading C45 model\n");
    
    /////// C4.5 initialization code ///////////////////////////// 
    GetNames();
    DecisionTree = GetTree(".tree");
    //printf("MaxAtt = %d, MaxItem = %d, MaxClass = %d\n",MaxAtt,MaxItem,MaxClass);
    RangeDesc = (struct ValRange *) calloc(MaxAtt+1, sizeof(struct ValRange));
    for(a=0; a<=MaxAtt; a++)
    {
        if ( MaxAttVal[a] )
        {
            RangeDesc[a].Probability =
                    (float *) calloc(MaxAttVal[a]+1, sizeof(float));
        }
    }
    ClassSum = (float *) malloc((MaxClass+1) * sizeof(float));
    LowClassSum = (float *) malloc((MaxClass+1) * sizeof(float));
    ///////////////////////////////////////////////////////////////
    
    
    numFeatures = MaxAtt + 1;
    numClasses = MaxClass + 1;
    inBuffer = (float *) malloc(numFeatures * sizeof(float));
    outBuffer = (double *) malloc(numClasses * sizeof(double));
    
    if (inSignal) {
        if (!SignalConnectWait(inSignal)){
            printf("C45Classifier: Error connecting to %s\n",inUNL);
            exit(1);
        }
        if (SignalGetLength(inSignal, 1) != numFeatures)
        {
            printf("C45Classifier: Error, numFeatures (%d) does not match length of input signal (%d)\n",
                   numFeatures, SignalGetLength(inSignal,1));
            exit(1);
        }
    }
    /*if (postUNL != NULL) {
        probabilityVector = NewVector(numClasses);
        nodeName = strrchr(postUNL, '/');
        if(nodeName == NULL) {
            printf("C45Classifier: Error extracting node name from partial unl %s\n",postUNL);
            exit(1);
        }
        nodeName++;
        updateNode = CreateVectorNode(nodeName, T_VECTOR, probabilityVector);
    }*/
    
    //printf("%d classes, %d feature vector\n",numClasses,numFeatures);
    
    if (outUNL != NULL) {
        outSignal = NewSignalGenUNL(outUNL);
        if(outSignal == NULL) {
            printf("C45Classifier: Error allocating outSignal %s\n",outUNL);
            exit(1);
        }
        SignalSetType(outSignal, T_DOUBLE);
        SignalSetDimension(outSignal, 1, numClasses, NULL, NULL);
        printf("C45Classifier: sending signal on %s\n",outUNL);
    }
    
    if (outFile != NULL) {
        
        if (strrchr(outFile,'/') != NULL) {
            //extract path and create directory
            sprintf(temp,"mkdir -p ");
            n = (int)(strrchr(outFile,'/')-outFile);
            strncpy (temp+9,outFile,n);
            temp[n+9]='\0';
            system(temp);
        }
        outFile_fid = fopen(outFile,"w");
        if (outFile_fid == NULL) {
            printf("C45Classifier: Error, could not create file file %s\n", outFile);
            exit(1);
        }
    }

    printf("C45Classifier: Reading Signals -----------\n");
    while(runflag) {
        if (inSignal != NULL) {
            if (SignalRecv(inSignal, &time, inBuffer) != 1) {
                continue;  //SignalRecv did not recieve new data, so do not compute classification
            }
        }
        else // (inFile_fid != NULL)
        {
            if (fileTimeStamps) {
                if (fscanf(inFile_fid,"%ld.%ld",&time.tv_sec,&time.tv_usec) == EOF) {
                    printf("C45Classifier: Reached end of input file\n");
                    break;  //this is the normal way to terminate since we end after a newLine
                }
            }
            for (i = 0; i<numFeatures; i++) {
                if (fscanf(inFile_fid,"%f",&inBuffer[i]) == EOF) {
                    if (fileTimeStamps){
                        printf("C45Classifier: Warning, unfinished feature vector at end of input file\n");
                    }
                    else {
                        printf("C45Classifier: Reached end of input file\n");
                    }
                    endOfFile = 1;
                    break;  //no runtime error, but the last classification will not be correct
                }
            }
            if (endOfFile) break;
            //fgets(inBuffer,numFeatures * sizeof(float),inFile_fid);
        }

          
        ////// C4.5 code that does the actual classification //////
        for(a=0; a<=MaxAtt; a++) {
            RangeDesc[a].Known = true;
            RangeDesc[a].Asked = true;
            RangeDesc[a].LowerBound = inBuffer[a];
            RangeDesc[a].UpperBound = inBuffer[a];
        }
                
        for(c=0; c<=MaxClass; c++)
        {
            LowClassSum[c] = ClassSum[c] = 0;
        }
       
        //  Find the likelihood of an item's being of each class
        ClassifyCase(DecisionTree, 1.0);  
 
        // Find the best class and show decision made  
        BestClass = 0;
        
        for(c=0; c<=MaxClass; c++)
        {
            //Verbosity(1) printf("class %d weight %.2f\n", c, ClassSum[c]);
            Uncertainty -= LowClassSum[c];
            if ( ClassSum[c] > ClassSum[BestClass] ) BestClass = c;
            outBuffer[c] = (double)ClassSum[c];
        }
        //printf("best class = %d entspr. %s\n",BestClass,ClassName[BestClass]);
        //Decision(BestClass, ClassSum[BestClass],LowClassSum[BestClass],Uncertainty + LowClassSum[BestClass]);
        
        ////////////////////////////////////////////////////////////

        if (useCurrentTime) {
            //printf("Sleeping\n");
            usleep(100000); //0.1sec
            gettimeofday(&time, NULL);
        }
        
        if (postUNL != NULL) {
            nodeName = strrchr(postUNL, '/');
            if(nodeName == NULL) {
                printf("C45Classifier: Error extracting node name from partial unl\n");
                return;
            }
            nodeName++;
            probabilityVector = NewVector(numClasses);
            for (i = 0; i<numClasses; i++)
                    VectorSet(probabilityVector, i, outBuffer[i]);
            
            if (detailedPost) {
                updateNode = CreateBranchNode(nodeName, T_BRANCH);
                for(i=0; i<numClasses; i++) {
                    labelNode = CreateFloatNode(ClassName[i], T_DOUBLE, outBuffer[i]);
                    InsertNode(updateNode, labelNode);
                }
                labelNode = CreateStringNode("decision", T_STRING, ClassName[BestClass]);
                InsertNode(updateNode, labelNode);
                labelNode = CreateVectorNode("probs", T_VECTOR, probabilityVector);
                InsertNode(updateNode, labelNode);
                //PrintNode(updateNode, NULL);
                //PublishNode(NULL, NULL, &pubInfo);
            }
            else
            {
                updateNode = CreateVectorNode(nodeName, T_VECTOR, probabilityVector);
            }

            if (!PublishNode(postUNL, updateNode, pubInfo))
                printf("C45Classifier: Error publishing node\n");
            if (updateNode != NULL) FreeNode(updateNode); // this also seems to free probabilityVector!
        }
        
        if (outFile_fid != NULL) {
            
            if (fileTimeStamps) fprintf(outFile_fid,"%06ld.%06ld", time.tv_sec, time.tv_usec);
            for (i=0; i<numClasses; i++) {
                fprintf(outFile_fid," %-30.30g",outBuffer[i]);
            }
            fprintf(outFile_fid,"\n");
        }
        
        if (verbose){
            for (i=0; i<numClasses; i++) {
                printf("%f ",outBuffer[i]);
            }
            printf("%s\n",ClassName[BestClass]);
        }
        
        // Important: SignalSend is placed at the end because it overwrites outBuffer!
        if (outSignal != NULL) {
            SignalGenAccept(outSignal);
            SignalSend(outSignal, &time, outBuffer);
        }
        
    }
    if (outFile_fid != NULL) {
        fflush(outFile_fid);
    }
    printf("C45Classifier: Freeing memory\n");
    //if (updateNode != NULL) FreeNode(updateNode); // this also seems to free probabilityVector!
    //if (probabilityVector != NULL) FreeVector(probabilityVector);
    if (inBuffer != NULL) free(inBuffer);
    if (outBuffer != NULL) free(outBuffer);
    if (ClassSum != NULL) free(ClassSum);
    if (LowClassSum != NULL) free(LowClassSum);
    printf("Exiting nicely\n");
    return 0;
}


