/*****************************************************************************
 *
 * Author: Steve Dunn, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2002-2005 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************
* Rewritten by Mark Blum (mark@media.mit.edu) in July 2005
* 
* This program implements a Bayes classifier with single Gaussian PDF.
* It provides the Enchantment interface to the Gaussian classes, takes care of
* reading feature signals, calling classification functions, and posting the 
* results as signals or nodes on the whiteboard. File I/O is also supported
* for offline use.
*
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

extern "C" {
    #include <enchantment/signals.h>
    #include <enchantment/vector.h>
    #include <enchantment/whiteboard.h>
}

#include "Classification.hpp"
#include "Gaussian.hpp"
#include "MixtureModel.hpp"

//Static so we can flush it on
FILE* outFile_fid;
volatile int runflag = 1;
volatile int killflag = 0;

void PrintUsage()
{
    printf("Usage: GClassifier [OPTION]...\n");
    printf("                                                  \n");
    printf("       -m [FILE]     XML model of GMM \n");
    printf("       -i [UNL]      read input signal from [UNL] \n");
    printf("       -o [UNL]      write output signal to [UNL] \n");
    printf("       -p [UNL]      post output to [UNL] on whiteboard\n");
    printf("       -r [FILE]     read input from [FILE] \n");
    printf("       -w [FILE]     write output to [FILE] \n");
    printf("       -c [char*]    Context Label e.g. Location \n");
    printf("       -v            verbose, write output to stdout \n");
    printf("       -t            use current system time for output timestamps\n");
    printf("       In case of whiteboard posting: \n");
    printf("       -d            detailed post, includes a node per class with its\n");
    printf("                     probability, a node with probability vector and a \n");
    printf("                     decision node with a string of the most likely class.\n");
    printf("                     Per default, only the vector is posted.\n");
    printf("                                            \n");
}

void SigHandler(int sig) {
    switch (sig) {
        case SIGINT:
            runflag=0;
            if(killflag)
            {
                fflush(stdout);
                if (outFile_fid!=NULL){
                    fflush(outFile_fid);
                }
                fprintf(stderr,"Hard kill, flushing output streams and exiting.\n");
                exit(0);
            }
            else
            {
                killflag = 1;
            }
            
            break;
            default:
                printf("Signal handler received %d\n",sig);
    }
    return;
}

int main(int argc, char *argv[]) {
    FILE* inFile_fid;
    char *inUNL = NULL; // (ex. /signals/feat)
    char *outUNL = NULL;   // (ex. /classifiers/posture)
    char *postUNL = NULL;   // (ex. /classifiers/posture)
    char *inFile = NULL;
    char *outFile = NULL;
    char *modelFile = NULL;
    char *contextLabel = NULL;
    bool verbose = false;
    bool useCurrentTime = false;
    bool detailedPost = false;
    float* inBuffer = NULL;
    double* outBuffer = NULL;
    int numFeatures, numClasses;
    int endOfFile = 0;
    
    Signal *inSignal = NULL;
    Signal *outSignal = NULL;
    struct timeval time;
    //these 4 would take care of publishing a vector
    //char *nodeName;
    //NodeType *updateNode = NULL;
    //PublishNodeInfo *pubInfo = NULL;
    //Vector *probabilityVector = NULL;
    Vector *featureVector = NULL;
    Classification *classification;
    MixtureModel mixtureModel;
    int i;
    
    signal(SIGINT,SigHandler);  //die nicely
    printf("GClassifier start\n");
    
    ///// Parse input arguments ///////////////////////////////////////////
    
    i = 1;
    while ( i < argc ) {
        if (argv[i][0] == '-' ) {
            switch ( argv[i][1] ) {  // options without arguments
                case 'v': {
                    verbose = true;
                    break;
                }
                case 't': {
                    useCurrentTime = true;
                    break;
                }
                case 'd': {
                    detailedPost = true;
                    break;
                }
                default: {
                    i++;
                    if ( i == argc ) // check if argument follows
                    {
                        PrintUsage();
                        exit(1);
                    }
                    switch ( argv[i-1][1] ) { // options with arguments
                        case 'i': {
                            inUNL = argv[i];
                            break;
                        }
                        case 'o': {
                            outUNL = argv[i];
                            break;
                        }
                        case 'p': {
                            postUNL = argv[i];
                            break;
                        }
                        case 'r': {
                            inFile = argv[i];
                            break;
                        }
                        case 'w': {
                            outFile = argv[i];
                            break;
                        }
                        case 'm': {
                            modelFile = argv[i];
                            break;
                        }
                        case 'c': {
                            contextLabel = argv[i];
                            break;
                        }
                        default: {
                            printf("Unknown option -%c\n\n",argv[i-1][1]);
                            PrintUsage();
                            exit(1);
                        }
                    } // end options with arguments
                }
            } //end switch ( argv[i][1]
            i++;
        }
        else  // no '-' detected
        {
            PrintUsage();
            exit(1);
        }
    }  ///// end parsing input arguments
       
    if (inUNL==NULL && inFile==NULL){
        printf("GClassifier: Error, no input specified!\n\n");
        PrintUsage(); 
        exit(1);
    }
    if (modelFile==NULL){
        printf("GClassifier: Error, no model file specified!\n\n");
        PrintUsage(); 
        exit(1);
    }
    
    if (inUNL!=NULL) {
        inSignal = NewSignalUNL(inUNL);
        if(inSignal == NULL) {
            printf("GClassifier: Error allocating signal %s\n",inUNL);
            exit(1);
        }
    }
    else
    {
        inFile_fid = fopen(inFile,"r");
        if (inFile_fid == NULL) {
            printf("GClassifier: Error, could not open file %s\n", inFile);
            exit(1);
        }
    }
    
    printf("GClassifier: Loading mixture model\n");
    mixtureModel.loadXML(modelFile);
    if (verbose) mixtureModel.show();
  
    classification = new Classification(contextLabel, postUNL);
    numFeatures = mixtureModel.getDim();
    if (inSignal) {
        if (!SignalConnectWait(inSignal)){
            printf("GClassifier: Error connecting to %s\n",inUNL);
            exit(1);
        }
        if (SignalGetLength(inSignal, 1) != numFeatures)
        {
            printf("GClassifier: Error, numFeatures (%d) does not match length of input signal (%d)\n",
                   numFeatures, SignalGetLength(inSignal,1));
            exit(1);
        }
    }
    inBuffer = (float *) malloc(numFeatures * sizeof(float));
    featureVector = NewVector(numFeatures);
    numClasses = mixtureModel.getCount();
    outBuffer = (double *) malloc(numClasses * sizeof(double));
    /*if (postUNL != NULL) {
        probabilityVector = NewVector(numClasses);
        nodeName = strrchr(postUNL, '/');
        if(nodeName == NULL) {
            printf("GClassifier: Error extracting node name from partial unl %s\n",postUNL);
            exit(1);
        }
        nodeName++;
        updateNode = CreateVectorNode(nodeName, T_VECTOR, probabilityVector);
    }*/
    
    printf("%d classes, %d feature vector\n",numClasses,numFeatures);
    
    if (outUNL != NULL) {
        outSignal = NewSignalGenUNL(outUNL);
        if(outSignal == NULL) {
            printf("GClassifier: Error allocating outSignal %s\n",outUNL);
            exit(1);
        }
        SignalSetType(outSignal, T_DOUBLE);
        SignalSetDimension(outSignal, 1, numClasses, NULL, NULL);
    }
    
    if (outFile != NULL) {
        
        if (strrchr(outFile,'/') != NULL) {
            //extract path and create directory
            char temp[strlen(outFile)+10];
            sprintf(temp,"mkdir -p ");
            int n = (int)(strrchr(outFile,'/')-outFile);
            strncpy (temp+9,outFile,n);
            temp[n+9]='\0';
            system(temp);
        }
        outFile_fid = fopen(outFile,"w");
        if (outFile_fid == NULL) {
            printf("GClassifier: Error, could not create file file %s\n", outFile);
            exit(1);
        }
    }

    printf("GClassifier: Reading Signals -----------\n");
    while(runflag) {
        if (inSignal != NULL) {
            if (SignalRecv(inSignal, &time, inBuffer) != 1) {
                continue;  //SignalRecv did not recieve new data, so do not compute classification
            }
        }
        else // (inFile_fid != NULL)
        {
            if (fscanf(inFile_fid,"%ld.%ld",&time.tv_sec,&time.tv_usec) == EOF) {
                printf("GClassifier: Reached end of input file\n");  
                break;  //this is the normal way to terminate since we end after a newLine
            }
            else {
                for (i = 0; i<numFeatures; i++) {
                    if (fscanf(inFile_fid,"%f",&inBuffer[i]) == EOF) {
                        printf("GClassifier: Warning, unfinished feature vector at end of input file\n");  
                        endOfFile = 1;
                        break;  //no runtime error, but the last classification will not be correct
                    }
                }
            }
            if (endOfFile) break;
            //fgets(inBuffer,numFeatures * sizeof(float),inFile_fid);
        }
        for (i = 0; i<numFeatures; i++)
            VectorSet(featureVector, i,(double)(inBuffer[i]));
        //printf("received new feature Vector: ");
        //PrintVector(stdout,featureVector);
        
        mixtureModel.classify(classification, featureVector);
        //classification->show();
        classification->getProbabilities(&outBuffer[0]);
                        
        if (useCurrentTime) {
            //printf("Sleeping\n");
            usleep(100000); //0.1sec
            gettimeofday(&time, NULL);
        }
        
        if (postUNL != NULL) {
            //
            if (detailedPost) {
                classification->publish(postUNL);
            }
            else
            {
                classification->publishVector(postUNL);
            }
        }
        
        if (outFile_fid != NULL) {
            
            fprintf(outFile_fid,"%06ld.%06ld", time.tv_sec, time.tv_usec);
            for (i=0; i<numClasses; i++) {
                fprintf(outFile_fid," %-30.30g",outBuffer[i]);
            }
            //fprintf(outFile_fid," %s\n",classification->mostLikely());
            fprintf(outFile_fid,"\n");
        }
        
        if (verbose){
            for (i=0; i<numClasses; i++) {
                printf("%g ",outBuffer[i]);
            }
            printf(" %s\n",classification->mostLikely());
        }
        
        // Important: SignalSend is placed at the end because it overwrites outBuffer!
        if (outSignal != NULL) {
            SignalGenAccept(outSignal);
            SignalSend(outSignal, &time, outBuffer);
        }
        
    }
    if (outFile_fid != NULL) {
        fflush(outFile_fid);
    }
    printf("GClassifier: Freeing memory\n");
    //if (updateNode != NULL) FreeNode(updateNode); // this also seems to free probabilityVector!
    ////if (probabilityVector != NULL) FreeVector(probabilityVector);
    if (featureVector != NULL) FreeVector(featureVector);
    if (inBuffer != NULL) free(inBuffer);
    if (outBuffer != NULL) free(outBuffer);
    delete classification;
    printf("Exiting nicely\n");
    return 0;
}
