/*****************************************************************************
*
* Author: Steve Dunn, MIT Wearable Computing Group
* http://www.media.mit.edu/wearables/
* Copyright (C) 2002 MIT Media Lab
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*****************************************************************************
* Modified by Mark Blum (mark@media.mit.edu) in July 2005
* Modified by Jon Gips (jgips@media.mit.edu) in August 2005
*
* The Gaussian class is used by MixtureModel. It models a Gaussian probability
* density function. Includes XML parsing
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <assert.h>

#include "Gaussian.hpp"

Gaussian::Gaussian(unsigned int dim) 
  : meanXML(NULL), covXML(NULL), invcovXML(NULL)
{
  identifier = 0;
  label = NULL;
  dimension = dim;
  sigma_det = 0.0;
  u = NULL;
  cov = NULL;
  sigma_inv = NULL;
  
  _Verbose = 0;
}

Gaussian::~Gaussian() {
  clear();
}

double Gaussian::calc(Vector *feature) {
  Vector *feature_norm = NewVector(dimension) ;
    Vector *result = NewVector(dimension);
    double coeff, sum;
    double data = 0;;
    unsigned int i,j;

    if(feature->dimension != dimension ||
       u->dimension != dimension ||
       sigma_inv->rows != dimension ||
       sigma_inv->cols != dimension) {
        printf("Error - invalid vector and matrix dimensions do not match feature vector dimensions\n");
        printf("dimension: %d RealVector: %d u: %d sigma_inv rows: %d sigma_inv cols: %d\n",
               dimension, feature->dimension, u->dimension, sigma_inv->rows, sigma_inv->cols);
        return 0;
       }

    // coeff = 1.0 / sqrt( pow(2*M_PI,dimension) * sigma_det);
    // prob = coeff * exp( -0.5 * (X-u) . sigma_inv . (X-u) );

       if(_Verbose) {
           printf("Label: %s\n", label);
           printf("Det: %e\n", sigma_det);
       }
       coeff = 1.0 / sqrt( pow(2*M_PI,dimension) * sigma_det);
       if(_Verbose)
           printf("Coeff: %e\n", coeff);
       VectorSubtractStat(feature_norm, feature, u);
       if(_Verbose) {
           printf("Subtract\n");
           PrintVector(stdout,feature_norm);
       }
       //RealVector::multiply(&feature_norm, sigma_inv, &result);
       
       // multiplication ///////////////////////
       for(i=0; i<dimension; i++) {
           sum = 0;
           for(j=0; j<dimension; j++) {
               sum += VectorRef(feature_norm,j)*MatrixRef(sigma_inv,j,i);
           }
           VectorSet(result, i, sum);
       }
       // end multiplication ///////////////////////
       if(_Verbose) {
           printf("Multiply\n");
           PrintVector(stdout,result);
       }
       //data = RealVector::innerProduct(&result, &feature_norm);
       for(i=0; i<dimension; i++)
           data += VectorRef(result,i)*VectorRef(feature_norm,i);
       
       if(_Verbose)
           printf("innerProduct: %e\n", data);

       data *= -0.5;
       if(_Verbose)
           printf("-0.5*innerProduct: %e\n", data );
       data = exp( data );
       if(_Verbose)
           printf("exp(...): %e\n", data );
       data = coeff * data;
       if(_Verbose)
           printf("Final Prob: %e\n\n", data);

       FreeVector(feature_norm);
       FreeVector(result);
       return data;
}

void Gaussian::show() {
    //unsigned int i,j;
    printf("Gaussian:\n");
    printf("Label: %s\n", label);
    printf("Dimension: %d\n", dimension);
    printf("Sigma_det: %e\n", sigma_det);
    printf("u:\n");
    if(u != NULL)
        PrintVector(stdout,u);
    else
        printf(" Empty\n");
    printf("Sigma_inv:\n");
    if(sigma_inv != NULL)
      PrintMatrix(stdout,sigma_inv);
    else if(cov != NULL) 
      PrintMatrix(stdout,cov);
    else
      printf(" Empty\n");
    
//     printf("u:");
//     if(u != NULL){
//         for(i=0;i<dimension; i++)
//             printf(" %f",u[i]);
//         printf("\n");
//     }
//     else
//         printf(" Empty\n");
//     printf("Sigma_inv:");
//     if(sigma_inv != NULL){
//         for(i=0;i<dimension; i++){
//             for(j=0;j<dimension; j++)
//                 printf(" %f",sigma_inv[i][j]);
//             printf("\n");
//         }
//     }
//     else
//         printf(" Empty\n");
}



void Gaussian::clear(void)
{
  //do not clear dimension as it is only passed in constructor!
  if (u != NULL) FreeVector(u);
  if (sigma_inv != NULL) FreeMatrix(sigma_inv);
  if (label != NULL) free(label);
  if (invcovXML != NULL) delete invcovXML;
  if (meanXML != NULL) delete meanXML;

  u = NULL;
  sigma_inv = NULL;
  label = NULL;
  meanXML = NULL;
  invcovXML = NULL;
}

//--------------------------------------------------------------------

void Gaussian::startXMLElement(const char *name, const char **atts) 
{
  if(!strcasecmp(name, "gaussian")) {
    xml_state = GAUSSIAN;
    clear();
  }
  else if(!strcasecmp(name, "label")) {
    xml_state = LABEL;
  } 
  else if(!strcasecmp(name, "det")) {
    xml_state = DET;
  } 
  else if(!strcasecmp(name, "mean")) {
    xml_state = MEAN;
    meanXML = new XMLVector(dimension);
    pushXMLObject(meanXML, name, atts);
  }
  else if(!strcasecmp(name, "covariance")) {
    xml_state = COV;
    covXML = new XMLMatrix(dimension, dimension);
    pushXMLObject(covXML, name, atts);
  } 
  else if(!strcasecmp(name, "inversecovariance")) {
    xml_state = INVCOV;
    invcovXML = new XMLMatrix(dimension, dimension);
    pushXMLObject(invcovXML, name, atts);
  } 
  else {
    printf("Got unknown start tag \"%s\".\n", name);
    assert(0);
  }
}

void Gaussian::processXMLCharacterData(const char *s, int len)
{
  switch(xml_state) {
  case LABEL:  
    label = MyStrndup(s,len);
    break;
  case DET:  
    sigma_det = atof(s);
    break;
  default:
    break;
  }
  xml_state = SKIP; //only one parsing necessary per state
} 

void Gaussian::endXMLElement(const char *name)
{
  if(!strcasecmp(name, "gaussian")) {
    if(meanXML) {
      u = meanXML->v;
      meanXML->v = NULL;
      delete meanXML;
      meanXML = NULL;
    }
    
    if(covXML) {
      cov = covXML->m;
      covXML->m = NULL;
      delete covXML;
      covXML = NULL;
    }

    if(invcovXML) {
      sigma_inv = invcovXML->m;
      invcovXML->m = NULL;
      delete invcovXML;
      invcovXML = NULL;
    }
    
    popXMLObject();
  }
}
