/*************************************************
 * File: types.h
 * Birth Date: Jul 28, 2004
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 Aware Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 *************************************************/

#ifndef ENCHANTMENT_TYPES_H
#define ENCHANTMENT_TYPES_H

#include <enchantment/darray.h>

#include <enchantment/vector.h>
#include <enchantment/matrix.h>

#include <time.h>
#include <stdint.h>

typedef uint32_t TypeID;

#define T_NONE				(0x00000000)

#define FORMAT_MASK			  (0xFF0000)

#define T_EMPTY				  (0x010000)
#define T_NUMERIC			  (0x020000)
#define T_STRING			  (0x030000)
#define T_TYPE				  (0x040000)
#define T_TIMEVAL			  (0x050000)
#define T_NETSOCK			  (0x060000)
#define T_BRANCH			  (0x070000)
#define T_BOOL				  (0x080000)
#define T_CHARACTER			  (0x090000)

#define T_UNKNOWN			  (0x0A0000)

// Empty types

// Handle to something
#define T_HANDLE		(T_EMPTY |   0x0100)
// Handle to a cert
#define T_CERTIFICATE		(T_HANDLE |    0x01)

// There's a value, but we didn't get it
#define T_OPAQUE		(T_EMPTY |   0x0200)

// Numeric types

#define T_NUMERIC_FORMAT_MASK             (0x000F80)

// Signed integer of some size
#define T_INT_FORMAT		            (0x0080)

// IEEE floating point of some size
#define T_FLOAT_FORMAT	       		    (0x0180)

#define T_INT8      (T_NUMERIC | T_INT_FORMAT |   0)
#define T_INT16	    (T_NUMERIC | T_INT_FORMAT |   1)
#define T_INT32	    (T_NUMERIC | T_INT_FORMAT |   2)

#define T_FLOAT	    (T_NUMERIC | T_FLOAT_FORMAT | 2)
#define T_DOUBLE    (T_NUMERIC | T_FLOAT_FORMAT | 3)

#define T_FIXED_FORMAT			    (0x0280)

#define T_FRACTION_FORMAT		    (0x0380)

// Strings
#define T_UNL		(T_STRING |	     0x0100)
#define T_AUTH		(T_STRING |	     0x0200)

// Netsocks
#define T_SIGNAL	(T_NETSOCK |	     0x0001)
#define T_SERVER	(T_NETSOCK |	     0x0002)

// Tensors
// Vector
#define T_VECTOR   (T_DOUBLE + 0x1000)
// Matrix
#define T_MATRIX   (T_DOUBLE + 0x2000)

#define GetFormat(x)            ((x) & FORMAT_MASK)
#define HasFormat(x, format)	(GetFormat(x) == (format))

// values of type T_NETSOCK
typedef struct 
{
  uint16_t port; // the target port in network byte order
  uint32_t ip; // the target ip address in network, ready to connect
} NetSocketLocation;

NetSocketLocation *CreateNetSocketLocation(const char *hostname, 
					   unsigned port);
NetSocketLocation *CopyNetSocketLocation(NetSocketLocation *orig);

typedef union 
{
  void *empty_val; // used to zero out the node's data field

  /* 
   * used for all integer types which fit in 32 bits and also
   * characters (as UCS-32) 
   */
  int int_val;

  float float_val;
  char *string_val;
  double *doubleptr_val;
  NetSocketLocation *netsock_val;
  struct timeval *timeval_val;
  DArray branch_val;
  TypeID type_val;
  Vector *vector_val;
  Matrix *matrix_val;
} EnchantData;

/* 
 * Returns a human-readable string name of the given type, or NULL if
 * it is not a defined type. 
 */
const char *TypeName(TypeID type);

/* Returns the type with a given string name. */
TypeID TypeByName(const char *name);

void DataFromNumeric(EnchantData *data, TypeID dataType, 
		     void *val, TypeID valType);

void NumericFromData(void *val, TypeID valType, 
		     EnchantData *data, TypeID dataType);



#endif
