/*************************************************
 * File: XSpectrumViewer.c
 * Author: Daniel Barkalow
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *************************************************/

#include <enchantment/signals.h>
#include <enchantment/time-utils.h>

#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <unistd.h>

#include <string.h>

#include <math.h>

#include "XViewer.h"

int width;
int height;

int step = 1;

int shift;
int factor;

int max_color = 1024;

void set_range(double min, double max)
{
  factor = (int) ((double) max_color / (max - min));
  shift = (int) -min;
  printf("max %g, min %g\n", max, min);
  //printf("shift %d factor %d\n", shift, factor);
}

unsigned long mono_color(double value)
{
  int idx = (value + shift) * factor;
  if (idx > 1023)
    idx = 1023;
  if (idx < 0)
    idx = 0;
  return ((idx / 4) * (1 + (1 << 8) + (1 << 16)));
}

unsigned long old_color(double value)
{
  int idx = (value + shift) * factor;
  if (idx < 0)
    return 0x000000;
  else if (idx < 256)
    return ((idx - 0) << 0);
  else if (idx < 512)
    return ((idx - 256) << 8) + ((511 - idx) << 0);
  else if (idx < 768)
    return ((idx - 512) << 16) + ((767 - idx) << 8);
  else if (idx < 1024)
    return 0xFF0000 + ((idx - 768) * ((1 << 8) + (1 << 0)));
  else
    return 0xFFFFFF;
}

unsigned long jet_color(double value)
{
  int idx = (value + shift) * factor;

  if (idx < 0)
    return 0x00007F;
  else if (idx < 128)
    return 0x00007F + ((idx - 0) << 0);
  else if (idx < 384)
    return 0x0000FF + ((idx - 128) << 8);
  else if (idx < 640)
    return 0x00FF00 + ((639 - idx) << 0) + ((idx - 640) << 16);
  else if (idx < 896)
    return 0xFF0000 + ((895 - idx) << 8);
  else if (idx < 1024)
    return 0x7F0000 + ((1023 - idx) << 16);
  else
    return 0x7F0000;
}

unsigned long wheel_color(double value)
{
  int idx = (value + shift) * factor * 3 / 4;

  while (idx < 0)
    idx += 768;
  while (idx >= 768)
    idx -= 768;
  if (idx < 256)
    return ((255 - idx) << 0) + ((idx - 0) << 8);
  else if (idx < 512)
    return ((511 - idx) << 8) + ((idx - 256) << 16);
  else
    return ((767 - idx) << 16) + ((idx - 512) << 0);
}

// XXXX we draw the blocks from the previous timestamp to the current
// timestamp, rather than centered around the current timestamp

int main(int argc, char **argv)
{
  int above[2]; // count above max, 2max+1min/3
  int below[2]; // count below min, 2min+1max/3

  int count = 0;
  char *UNL = argv[1];
  float *buffer = NULL;
  int *last;
  char opt;
  int debug = 0;
  Signal *signal;
  char *title = NULL;
  long timestep = 20000l;
  double max, min;
  int y, fy;
  int autoscale = 1;
  int length;
  int root = 0;

  unsigned long (*color)(double);

  double maxopt, minopt;

  XGCValues vals;
  XViewer *viewer;
  XImage *im;
  char *imdata;
  int proxy = 0;
  int reset = 0;
  int scaletime = 1;
  float rate;

  int logscale = 1;

  long lastX = -1;
  struct timeval prevtime = {0, 0};

  width = 500;
  height = 512;

  minopt = 1e-10;
  maxopt = 1e+4;

  above[0] = above[1] = below[1] = below[0] = 0;

  color = jet_color;

  do
    {
      opt = getopt(argc, argv, "nh:l:dt:ps:rH:c:");
      if (opt == 'n')
	logscale = 0;
      if (opt == 'h')
	{
	  maxopt = atof(optarg);
	  autoscale = 0;
	}
      if (opt == 'l')
	{
	  minopt = atof(optarg);
	  autoscale = 0;
	}
      if (opt == 'r')
	root = 1;
      if (opt == 't')
	title = optarg;
      if (opt == 'd')
	debug = 1;
      if (opt == 'p')
	proxy = 1;
      if (opt == 's')
	timestep = atoi(optarg);
      if (opt == 'H')
	height = atoi(optarg);
      if (opt == 'c')
	{
	  if (!strcmp(optarg, "wheel"))
	    color = wheel_color;
	  else if (!strcmp(optarg, "jet"))
	    color = jet_color;
	  else if (!strcmp(optarg, "mono"))
	    color = mono_color;
	  else
	    {
	      fprintf(stderr, "Unknown color scheme: %s\n", optarg);
	    }
	}
    }
  while (opt > 0);

  viewer = make_XViewer(NULL, root, title, timestep, width, height);

  if (logscale)
    {
      min = log10(minopt);
      max = log10(maxopt);
    }
  else
    {
      min = minopt;
      max = maxopt;
    }

  set_range(min, max);

  if (argc <= optind)
    {
      fprintf(stderr, "Usage: %s [OPTION]... UNL\n", argv[0]);
      fprintf(stderr,
	      "Display signal spectra graphically as a color field.\n\n");
      fprintf(stderr, "  -h [VALUE]	start with [VALUE] as the max.\n");
      fprintf(stderr, "  -l [VALUE]	start with [VALUE] as the min.\n");
      fprintf(stderr, "  -t [NAME]	title the window [NAME].\n");
      fprintf(stderr, "  -p       	use the server as a proxy.\n");
      fprintf(stderr, "  -s [MSEC]	make the time step [MSEC].\n");
      fprintf(stderr, "\n");
      fprintf(stderr, 
	      "The default is %g to %g, with a time step of %ld usec.\n",
	      minopt, maxopt, timestep);
      fprintf(stderr, 
	      "The data is handled on a logarythmic scale.\n");
      exit(1);
    }  
  
  UNL = argv[optind];

  imdata = malloc(height * 4 * 1000);
  im = XCreateImage(viewer->d, DefaultVisual(viewer->d, 0), 24, ZPixmap, 0,
		    imdata, 1, height, 8, 4);
  XFlush(viewer->d);
  
  signal = NewSignalUNL(UNL);
  //signal->conveltsize = 8;

  printf("Viewing signal at %s\n", UNL);
  while (1)
    {
      struct timeval time;
      if (!SignalConnected(signal))
	{
	  if (proxy ? !SignalConnectByProxy(signal) : !SignalConnectWait(signal))
	    {
	      printf("Failed to wait for signal publication\n");
	      exit(1);
	    }
	  if (SignalConnected(signal))
	    {
	      printf("Connected\n");
	      if (buffer) 
		{
		  free(buffer);
		  free(last);
		}
	      SignalSetType(signal, T_FLOAT);
	      buffer = malloc(SignalBufferSize(signal));
	      last = malloc(SignalGetLength(signal, 1) * sizeof(int));
	      viewer->gcs = malloc(sizeof(GC));
	      viewer->gcs[0] = XCreateGC(viewer->d, viewer->win, 0, &vals);
	      reset = 1;
	      length = SignalGetLength(signal, 1);
	      rate = SignalGetSampleRate(signal);
	    }
	}
      else
	{
	  if (SignalRecv(signal, &time, buffer))
	    {
	      int i;
	      long x;
	      if (prevtime.tv_sec && 
		  timeval_sub(&prevtime, &time) > 0)
		{
		  printf("%ld.%06ld < %ld.%06ld\n", time.tv_sec, time.tv_usec,
			 prevtime.tv_sec, prevtime.tv_usec);
		  reset = 1;
		}
	      prevtime = time;
	      if (!reset &&
		  (viewer->basetime.tv_sec > time.tv_sec ||
		   (viewer->basetime.tv_sec == time.tv_sec &&
		    viewer->basetime.tv_usec > time.tv_usec)))
		reset = 1;
	      if (autoscale && count > scaletime * SignalGetLength(signal, 1))
		{
		  if (rescale(above, below, &count, &min, &max))
		    {
		      set_range(min, max);
		      scaletime = 1;
		    }
		  else
		    scaletime = 64;
		}
	      if (reset)
		{
		  reset = 0;
		  XViewer_reset(viewer, &time);
		}
	      //PrintSignalElements(stdout, signal, time, buffer);
	      //printf("%g offset\n", 1000000 / (rate / 2));

	      // The end of the block is after the timestamp by 1/(2 rate).
	      timeval_add(&time, 1000000 / rate / 2);

	      x = xpos(viewer, &time);
	      y = 0;
	      fy = 0;
	      for (i = 0; i < length; i++)
		{
		  double val;
		  unsigned long col;
		  int iy;
		  val = buffer[i];
		  if (logscale)
		    val = log10(val);
		  if (autoscale)
		    note_datum(above, below, &count, min, max, val);
		  col = color(val);
		  //printf("%6x ", col);
		  
		  iy = y;
		  fy += height;
		  y += fy / length;
		  fy = fy % length;
		  for (; iy < y; iy++)
		    {
		      /* Write the high frequencies (large i) at the top. */
		      XPutPixel(im, 0, height - iy - 1, col);
		    }
		}
	      //printf("\n");
	      draw_bar(viewer, im, lastX, &time);
	      lastX = x;
	      XFlush(viewer->d);
	    }
	  if (!SignalConnected(signal))
	    {
	      printf("Disconnected after %d samples read\n", count);
	    }
	}
    }
}
