/*************************************************
 * File: time-axis.c
 * Birth Date: Jan 11, 2004
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2004 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *************************************************/

#include "XViewer.h"

#include <enchantment/time-utils.h>

#include <stdio.h>
#include <stdlib.h>

// Modifies base such that it is 0 mod period ms.
static void syncTime(XViewer *viewer, struct timeval *time)
{
  viewer->basetime = *time;
  timeval_round(&viewer->basetime, viewer->timestep * viewer->width);
}

long xpos(XViewer *viewer, struct timeval *sample)
{
  long diffus = timeval_sub(sample, &viewer->basetime);
  return diffus / viewer->timestep;
}

static void clear(XViewer *viewer, struct timeval *sample)
{
  // clears from the last cleared spot to the sample + 5 pixels
  int x;
  x = xpos(viewer, sample);
  pthread_mutex_lock(&viewer->lock);
  if (viewer->clearX >= x)
    {
      pthread_mutex_unlock(&viewer->lock);
      return;
    }
  if (viewer->clearX < 0)
    viewer->clearX = x;
  if ((x % viewer->width) + 5 <= viewer->width &&
      viewer->clearX / viewer->width == x / viewer->width)
    {
      XClearArea(viewer->d, viewer->win, 
		 (viewer->clearX + 1) % viewer->width, 0, 
		 (x - viewer->clearX + 5), viewer->height, 0);
    }
  else
    {
      XClearArea(viewer->d, viewer->win, 
		 (viewer->clearX) % viewer->width + 1, 0, 
		 viewer->width - (viewer->clearX) % viewer->width,
		 viewer->height, 0);
      XClearArea(viewer->d, viewer->win, 0, 0, 
		 (x + 5) % viewer->width, 
		 viewer->height, 0);
    }
  viewer->clearX = x;
  pthread_mutex_unlock(&viewer->lock);
}

void XViewer_reset(XViewer *viewer, struct timeval *time)
{
  syncTime(viewer, time);
  viewer->clearX = -1;
}

void draw_mark(XViewer *viewer, int color, struct timeval *time)
{
  long x = xpos(viewer, time);
  XDrawLine(viewer->d, viewer->win, viewer->gcs[color], x % viewer->width, 0,
	    x % viewer->width, viewer->height);
}

void draw_line(XViewer *viewer, int color, int lastX, int lastY,
	       struct timeval *time, int y)
{
  int x = xpos(viewer, time);
  if (lastX < 0)
    return;
  clear(viewer, time);
  if ((lastX % viewer->width) <= (x % viewer->width))
    {
      // We're drawing left to right
      if (lastX <= x)
	{
	  // We're drawing forward
	  XDrawLine(viewer->d, viewer->win, viewer->gcs[color],
		    lastX % viewer->width, lastY,
		    x % viewer->width, y);
	}
      else
	{
	  // We're drawing backward?
	  XDrawLine(viewer->d, viewer->win, viewer->gcs[color],
		    lastX % viewer->width, lastY,
		    lastX % viewer->width, y);
	}
    }
  else
    {
      if (lastX < x)
	{
	  // We're drawing forward, but wrapped around the edge
	  XDrawLine(viewer->d, viewer->win, viewer->gcs[color],
		    lastX % viewer->width, lastY,
		    x % viewer->width + viewer->width, y);
	  XDrawLine(viewer->d, viewer->win, viewer->gcs[color],
		    lastX % viewer->width - viewer->width, lastY,
		    x % viewer->width, y);
	}
    }
}

void draw_bar(XViewer *viewer, XImage *im, int lastX, struct timeval *time)
{
  int px = lastX;
  int x = xpos(viewer, time);
  if (lastX < 0 || lastX > x)
    return;
  if (px < (int) (x - viewer->width))
    {
      px = x - viewer->width;
    }
  clear(viewer, time);
  for (; px < x; px++)
    {
      XPutImage(viewer->d, viewer->win, viewer->gcs[0], im, 0, 0,
		px % viewer->width, 0, 1, viewer->height);
    }
}

void note_datum(int *above, int *below, int *count, 
		double min, double max, double val)
{
  if (val < min)
    below[0]++;
  else if (val * 3 < min + min + max)
    below[1]++;
  else if (val * 3 < min + max + max)
    {
      below[1]++;
      above[1]++;
    }
  else if (val < max)
    above[1]++;
  else
    above[0]++;
  (*count)++;
}

int rescale(int *above, int *below, int *count, double *min, double *max)
{
  int ret = 0;
  int i;
  int maxd = 0;
  int mind = 0;
  if (above[0] * 10 > *count)
    maxd = 1;
    //    nmax = *max + (*max - *min) / 2;
  else if ((above[0] + above[1]) * 10 < *count)
    maxd = -1;
    //nmax = (*max + *min) / 2;

  if (below[0] * 10 > *count)
    mind = -1;
  //nmin = *min + (*min - *max) / 2;
  else if ((below[0] + below[1]) * 10 < *count)
    mind = 1;
  //nmin = (*max + *min) / 2;

  if (mind || maxd)
    {
      if (1)
	{
	  printf("%d%% above max, %d%% not above threshold\n",
		 above[0] * 100 / *count,
		 100 - (above[0] + above[1]) * 100 / *count);
	  printf("%d%% below min, %d%% not below threshold\n",
		 below[0] * 100 / *count,
		 100 - (below[0] + below[1]) * 100 / *count);
	}
      if (maxd == mind)
	{
	  // move the window
	  double step = (*max - *min) / 2 * maxd;
	  *min += step;
	  *max += step;
	}
      else if (maxd > 0 && mind < 0)
	{
	  double step = (*max - *min) / 2;
	  // double the window
	  *min -= step;
	  *max += step;
	}
      else if (maxd < 0 && mind > 0)
	{
	  double step = (*max - *min) / 4;
	  // halve the window
	  *min += step;
	  *max -= step;
	}
      else if (maxd > 0)
	{
	  double step = (*max - *min);
	  // double the window
	  *max += step;
	}
      else if (mind < 0)
	{
	  double step = (*max - *min);
	  // double the window
	  *min -= step;
	}
      else if (maxd < 0)
	{
	  double step = (*max - *min) / 2;
	  // halve the window
	  *max -= step;
	}
      else if (mind > 0)
	{
	  double step = (*max - *min) / 2;
	  // halve the window
	  *min += step;
	}
      ret = 1;
    }
  for (i = 0; i < 2; i++)
    {
      //printf("above: %d below: %d\n", above[i], below[i]);
      above[i] = 0;
      below[i] = 0;
    }
  *count = 0;
  return ret;
}

static void *viewer_thread(XViewer *viewer)
{
  while (1)
    {
      struct timespec;
      pthread_mutex_lock(&viewer->lock);
      pthread_cond_wait(&viewer->cond, &viewer->lock);
      pthread_mutex_unlock(&viewer->lock);
    }
  return NULL;
}

void setup_thread(XViewer *viewer)
{
  pthread_cond_init(&viewer->cond, NULL);
  pthread_mutex_init(&viewer->lock, NULL);
  pthread_create(&viewer->thread, NULL, 
		 (void *(*)(void *)) viewer_thread, viewer);
}

XViewer *make_XViewer(char *displayname, unsigned root, char *title,
		      unsigned long timestep, unsigned width, unsigned height)
{
  XViewer *ret = malloc(sizeof(XViewer));
  ret->d = XOpenDisplay(displayname);
  if (root)
    {
      ret->win = DefaultRootWindow(ret->d);
      ret->width = 1280;
      ret->height = 1024;
    }
  else
    {
      ret->win = XCreateSimpleWindow(ret->d, DefaultRootWindow(ret->d), 0,0,
				     width, height, 1, 0, 0);
      ret->width = width;
      ret->height = height;
      XMapWindow(ret->d, ret->win);
      if (title)
	XStoreName(ret->d, ret->win, title);
    }
  ret->timestep = timestep;
  return ret;
}
