/*******************************************************************
 * Name: Jonathan Gips
 * 
 * Copyright (C) 2002 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * -------------------
 * File: hashtable.c  
 * -----------------
 * This file implements the HashTable abstraction. It provides a very
 * acceptable solution to large storage containers with truly fast
 * access and searching.  The tradeoff is the hashtable's size, but
 * this is well worth the price, especially with today's capable 
 * computers.
 *
 * This implementation relies heavily on the DArray ADT.  It combines
 * the DArray's generic properties with an desirable storage 
 * implementation, and therefore they make a very nice combo.
 ********************************************************************/

#include "hashtable.h"
#include <enchantment/darray.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

/* define: DARRAY_LEN
 * ------------------
 * Set this #define to change the initial length of the bucket chains
 * used to implement the hash table. It can be no lower than one
 * and a large value will probably be inefficient if the client
 * is using an appropriate number of buckets.  DARRAY_LEN is also 
 * used as the initial value of "sort_len" when TableNew is called.
 * A later call to ChangeSortLen() will overwrite this value.
 */

#define DARRAY_LEN 4

/* HashImplementation
 * ------------------
 * Contains all the information relevant to a HashTable.  Some 
 * further information is kept a level deeper, in the embedded 
 * array DArrays which are dynamically allocated. It is worthy 
 * noting that the same hash, comparison, and freeing functions 
 * are used for all the data. Also, note that the buckets must
 * be allocated dynamically because the client may choose their
 * number.
 */

typedef struct HashImplementation {
  int elemSize;
  int nBuckets;
  
  TableHashFn hashFn;
  TableCompareFn compFn;
  TableElementFreeFn freeFn;

  DArray *buckets;
  int nElems;

  int sort_len;
} HashImplementation;

//bool ChangeSortLen(HashTable table, int newSortLen);

/* TableNew
 * --------
 * Initializes a HashTable for use. Several asserts may be raised
 * if the minimum parameters are not passed in.
 */

HashTable TableNew(int elemSize, int nBuckets, TableHashFn hashFn, 
		   TableCompareFn compFn, TableElementFreeFn freeFn) {
  HashTable table;
  int i;
  
  assert(elemSize > 0);
  assert(nBuckets > 0);
  assert(hashFn != NULL);
  assert(compFn != NULL);

  table = (HashImplementation*)malloc(sizeof(HashImplementation));

  table->elemSize = elemSize;
  table->nBuckets = nBuckets;

  table->hashFn = hashFn;
  table->compFn = compFn;
  table->freeFn = freeFn;
  
  table->buckets = (DArray*) malloc(sizeof(DArray)*nBuckets);

  for(i = 0; i < nBuckets; i++) table->buckets[i] = NULL;

  table->nElems = 0;

  table->sort_len = DARRAY_LEN;
 
  return table;
}

/* TableFree
 * --------
 * Free function that first calls the Client provided
 * free function by means of calling ArrayFree on each
 * of the buckets, if they have been allocated. This also
 * frees the darray bucket chains. Then, the array of 
 * buckets is freed, and finally the HashTableImplementation
 * is released.
 */

void TableFree(HashTable table) {
  int i;

  for(i = 0; i < table->nBuckets; i++)
    if(table->buckets[i] != NULL) ArrayFree(table->buckets[i]);

  free (table->buckets);
  free(table);
}

int TableCount(HashTable table) {return table->nElems;}

/* TableEnter
 * ----------
 * This is where all of the action happens in this ADT.
 * First, the hash function is called to get the appropriate
 * bucket.  If a bucket chain needs to be allocated, then it
 * is. This method of not using it till it is needed saves
 * space and time. If the bucket chain is already there, then
 * the equivalent entry is searched for.  If it is found, then
 * it is replaced.  If it is not, then it is added.
 */

void TableEnter(HashTable table, const void *newElem) {
  int hash_buck, equiv_index = NOT_FOUND;

  hash_buck = (*(table->hashFn))(newElem,table->nBuckets);

  if(table->buckets[hash_buck] == NULL) 
    table->buckets[hash_buck] = ArrayNew(table->elemSize, table->sort_len,
					 table->freeFn);
  
  else equiv_index = ArraySearch(table->buckets[hash_buck], newElem,
				 table->compFn, 0,
				 ArrayLength(table->buckets[hash_buck]) >=
				 table->sort_len);
  
  if(equiv_index == NOT_FOUND) {
    ArrayAppend(table->buckets[hash_buck], newElem);
    table->nElems++;
    if(ArrayLength(table->buckets[hash_buck]) >= table->sort_len) 
      ArraySort(table->buckets[hash_buck], table->compFn);
  }
  else ArrayReplaceAt(table->buckets[hash_buck],newElem,equiv_index);
}

/* TableLookup
 * -----------
 * Retrieves the element that corresponds to a key.
 * There is heavy use of the DArray ADT here.
 */

void *TableLookup(HashTable table, const void *elemKey) {
  int hash_buck, equiv_index;

  hash_buck = table->hashFn(elemKey, table->nBuckets);

  if(table->buckets[hash_buck] == NULL) {return NULL;}
 
  equiv_index = ArraySearch(table->buckets[hash_buck], elemKey,
			    table->compFn, 0,
			    ArrayLength(table->buckets[hash_buck]) >=
			    table->sort_len);
 
  
  if(equiv_index == NOT_FOUND) return NULL;
  else return ArrayNth(table->buckets[hash_buck], equiv_index);
}

/* TableMap
 * --------
 * Maps a certain function over the entire table by doing one
 * bucket chain (DArray) at a time.
 */

void TableMap(HashTable table, TableMapFn fn, void *clientData) {
  int i;

  assert(fn != NULL);
  
  for(i = 0; i < table->nBuckets; i++) {
    if(table->buckets[i] != NULL) ArrayMap(table->buckets[i], fn,
					   clientData);
  }
}

/* ChangeSortLen (experimental)
 * ----------------------------
 * This function is not in use, but it will allow me to change
 * the sorting length on the fly, as the hashtable grows and
 * efficiency considerations change.
 */

/*  
bool ChangeSortLen(HashTable table, int newSortLen) {
  if(newSortLen <= 0) return false;
  
  table->sort_len = newSortLen;
  return true;

  Possible Resorting of all the unsorted chains.
}
*/
