/*************************************************
 * File: signals.h
 * Author: Daniel Barkalow
 * Birth Date: 2/28/03
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *************************************************/

#ifndef ENCHANTMENT_SIGNALS_H
#define ENCHANTMENT_SIGNALS_H

#include <enchantment/node.h>
#include <enchantment/identity.h>

#include <sys/time.h>

#include <stdio.h>

typedef struct Signal Signal;

/* Prepares to generate a new signal, which will be published at the
 * specified UNL. 
 */
Signal *NewSignalGenUNL(char *UNL);

/* Prepares to generate a new signal, which will not be published. */
Signal *NewSignalGenNone();

/* Prepares to receive a signal at the address in the given node. */
Signal *NewSignalNode(NodeType *node);

/* Prepares to receive a signal which will be published at the given UNL. */ 
Signal *NewSignalUNL(char *UNL);

/* Sets the identity of the program sending or receiving the signal. */
void SignalSetIdentity(Signal *signal, Identity *identity);

/* Accept any clients trying to connect to this signal generator. */
void SignalGenAccept(Signal *signal);

/* Attempts to connect to a signal,but only if it is available right away. */
int SignalConnect(Signal *signal);

/* Attempts to connect to a signal, waiting if necessary for a signal
 * to be published at the given location. Returns FALSE if waiting
 * more would not help in connecting.
 */
int SignalConnectWait(Signal *signal);

/* Attempt to connect to the signal by way of the server that hosts
 * the UNL to the signal. SignalConnectWait() above also tries this
 * method if a direct connection could not be formed.
 */
int SignalConnectByProxy(Signal *signal);

/* Returns the number of bytes needed to hold a data unit for this signal. */
size_t SignalBufferSize(Signal *signal);

/* Sends a data unit from the given signal generator. Returns 1 on
 * success; only reports failure if certified delivery of data is
 * specified and the library cannot add this unit to its buffer.
 *
 * See doc/Signals for details.
 */
int SignalSend(Signal *signal, struct timeval time, void *buffer);

/* Receives a data unit from the given signal.
 *
 * May be used on a signal generator with certified delivery to take
 * data units for local storage. In this case, the unit is treated as
 * being successfully received.
 *
 * See doc/Signals for details.
 */
int SignalRecv(Signal *signal, struct timeval *time_p, void *buffer);

/* Requires that data sent to this signal be acknowledged by a client
 * with the specified auth token or committed to local storage before
 * it is discarded.
 */
void SignalGenGetReceipts(Signal *signal, char *auth);

/* Reports that data at the given time has been dealt with successfully. */
void SignalRecvd(Signal *signal, struct timeval time);

/* Sets the time to the most recent acknowledgement on this signal. */
void SignalSent(Signal *signal, struct timeval *time_p);

/* Disconnects a receiver from the signal, or closes a
 * generator. Deallocates the signal memory in either case. 
 */
void SignalClose(Signal *signal);

/* Copies all of the properties from one signal receiver or generator
 * to another. Will only work if the target can be modified to match.
 */
void SignalCopyProperties(Signal *source, Signal *target);

/* Sets the signal's type to the specified enchantment type, which must be
 * a numeric type. If this is called on the input side, it must be a type
 * compatible with the output.
 */
void SignalSetType(Signal *signal, TypeID type);

/* Returns the type for the signal. */
TypeID SignalGetType(Signal *signal);

/* Sets the length of the signal in a non-time dimension, the label
 * for that dimension, and the labels for indices in that
 * dimension. Dimensions are numbered starting with 1. Label may be
 * NULL, if the dimension should not have a label; indexLabels may be
 * NULL if the dimension should not have any.
 */
void SignalSetDimension(Signal *signal, int dimension, 
			unsigned length, char *label, char **indexLabels);

/* Returns the length of a signal in some dimension. */
int SignalGetLength(Signal *signal, int dimension);

/* Sets the sample rate of the signal (in samples/second) and the
 * number of samples in a data unit. 
 */
void SignalSetSampleRate(Signal *signal, unsigned hz, unsigned length);

/* Returns the sample rate of the signal (in samples/second). */
unsigned SignalGetSampleRate(Signal *signal);

/* Returns the number of samples in a data unit. */
int SignalGetUnitSamples(Signal *signal);

/* Returns the number of clients currently receiving data. */
int SignalClientCount(Signal *signal);

/* Returns the address that a signal is coming from. */
NetSocketLocation *SignalSource(Signal *signal);

/* Returns true if this is connected. */
int SignalConnected(Signal *signal);

/* Writes a data unit to a file. */
void PrintSignalElements(FILE *file, Signal *signal, struct timeval time,
			 char *buffer);

/* Reads a data unit from a file. If buffer is NULL, reads the signal
 * properties instead. */
int ScanSignalElements(FILE *file, Signal *signal,
		       struct timeval *time, char *buffer);

#endif
