/*************************************************
 * File: XSignalViewer.c
 * Author: Daniel Barkalow
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *************************************************/

#include <enchantment/signals.h>

#include <stdlib.h>

#include <X11/Xlib.h>
#include <getopt.h>

#include "XViewer.h"

double max, min;
int width = 500;
int height = 256;
int step = 10;
int istep = 6;

int ypos(int value)
{
  return (int)((max - value) * height / (max - min));
}

unsigned long color(int linenum)
{
  switch (linenum)
    {
    case 0:
      return 0xff0000;
    case 1:
      return 0x00ff00;
    case 2:
      return 0x0000ff;
    case 3:
      return 0xffff00;
    default:
      return 0xffffff;
    }
}

GC *getGCs(Display *d, Window win, unsigned number)
{
  XGCValues vals;
  GC *gc;
  int i;
  gc = malloc(number * sizeof(GC));
  for (i = 0; i < number; i++)
    {
      vals.foreground = color(i);
      gc[i] = XCreateGC(d, win, GCForeground, &vals);
    }
  return gc;
}

int main(int argc, char **argv)
{
  int above[2];
  int below[2];
  int count = 0;
  char *UNL = argv[1];
  char *buffer = NULL;
  int *last;
  char opt;
  int debug = 0;
  Signal *signal;
  char *title = NULL;
  int proxy = 0;
  int reset = 0;
  int scaletime = 1;
  int do_rescale = 0;

  long lastX = 0;
  unsigned long timestep = 20l;

  GC *gc;
  Display *d = XOpenDisplay(NULL);
  Window win;

  min = -128;
  max = 127;
  above[0] = above[1] = below[1] = below[0] = 0;

  do
    {
      opt = getopt(argc, argv, "h:l:dt:ps:");
      if (opt == 'h')
	max = atoi(optarg);
      if (opt == 'l')
	min = atoi(optarg);
      if (opt == 't')
	title = optarg;
      if (opt == 'd')
	debug = 1;
      if (opt == 'p')
	proxy = 1;
      if (opt == 's')
	timestep = atoi(optarg);
    }
  while (opt > 0);
  if (argc <= optind)
    {
      fprintf(stderr, "Usage: %s [OPTION]... UNL\n", argv[0]);
      fprintf(stderr,
	      "Display signal data graphically as a set of traces.\n\n");
      fprintf(stderr, "  -h [VALUE]	start with [VALUE] as the max.\n");
      fprintf(stderr, "  -l [VALUE]	start with [VALUE] as the min.\n");
      fprintf(stderr, "  -t [NAME]	title the window [NAME].\n");
      fprintf(stderr, "  -p       	use the server as a proxy.\n");
      fprintf(stderr, "  -s [MSEC]	make the time step [MSEC].\n");
      fprintf(stderr, "\n");
      fprintf(stderr, 
	      "The default is %g to %g, with a time step of %ld msec.\n",
	      min, max, timestep);
      exit(1);
    }
  printf("min: %g max: %g\n", min, max);
  UNL = argv[optind];

  win = XCreateSimpleWindow(d, DefaultRootWindow(d), 0,0, width, height, 1, 0,
			    0);
  XMapWindow(d, win);
  if (title)
    XStoreName(d, win, title);
  XFlush(d);
  
  signal = NewSignalUNL(UNL);

  printf("Viewing signal at %s\n", UNL);
  while (1)
    {
      struct timeval basetime;
      struct timeval time;
      if (!SignalConnected(signal))
	{
	  if (proxy ? !SignalConnectByProxy(signal) : !SignalConnectWait(signal))
	    {
	      printf("Failed to wait for signal publication\n");
	      exit(1);
	    }
	  if (SignalConnected(signal))
	    {
	      printf("Connected\n");
	      if (buffer) 
		{
		  free(buffer);
		  free(last);
		}
	      SignalSetType(signal, T_INT32);
	      buffer = malloc(SignalBufferSize(signal));
	      last = malloc(SignalGetLength(signal, 1) * sizeof(int));
	      gc = getGCs(d, win, SignalGetLength(signal, 1) ? : 1);
	      reset = 1;
	    }
	}
      else
	{
	  if (SignalRecv(signal, &time, buffer))
	    {
	      int x, i;
	      if (do_rescale && count > scaletime * SignalGetLength(signal, 1))
		{
		  if (rescale(above, below, &count, &min, &max))
		    {
		      printf("min: %g max: %g\n", min, max);
		      scaletime = 640;
		    }
		  else
		    scaletime = 640;
		}
	      if (reset)
		{
		  reset = 0;
		  basetime.tv_usec = time.tv_usec;
		  basetime.tv_sec = time.tv_sec;
		  syncTime(&basetime, timestep * width);
		  lastX = -1;
		}
	      x = xpos(&basetime, &time, timestep);
	      clear(d, win, lastX, x, width, height);
	      if (!SignalGetLength(signal, 1))
		{
		  XDrawLine(d, win, gc[0], x % width, 0,
			    x % width, height);
		}
	      for (i = 0; i < SignalGetLength(signal, 1); i++)
		{
		  int y;
		  int val;
		  val = (((int *)buffer)[i]);
		  y = ypos(val);
		  note_datum(above, below, &count, min, max, val);
		  if ((lastX % width) <= (x % width))
		    {
		      if (lastX < 0)
			;
		      else 
			{ 
			  if (lastX <= x)
			    XDrawLine(d, win, gc[i], lastX % width, last[i],
				      x % width, y);
			  else
			    XDrawLine(d, win, gc[i], lastX % width, last[i],
				      lastX % width, y);
			}
		    }
		  else
		    {
		      if (lastX < x)
			{
			  XDrawLine(d, win, gc[i], lastX % width, last[i],
				    x % width + width, y);
			  XDrawLine(d, win, gc[i], 
				    lastX % width - width, last[i],
				    x % width, y);
			}
		    }
		  if (debug)
		    {
		      printf("%d->%d ", ((int *)buffer)[i], y);
		      if (i + 1 == SignalGetLength(signal, 1))
			printf("\n");
		    }
		  //printf("%d,%d to %d,%d\n", x, last[i], x + step, y);
		  last[i] = y;
		}
	      if (lastX < x)
		lastX = x;
	      XFlush(d);
	      count++;
	    }
	}
    }
}
