/*************************************************
 * File: signals.c
 * Author: Daniel Barkalow
 * Birth Date: 2/28/03
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002, 2003, 2004 MIT Media Lab, 
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *************************************************/

#include <enchantment/signals.h>
#include <enchantment/types.h>
#include <enchantment/whiteboard.h>
#include <stdlib.h>
#include <sys/param.h>
#include <string.h>
#include <netinet/in.h>

#include <enchantment/identity.h>
#include "protocol.h"
#include "introduction.h"
#include "ipc.h"

#define MAX_SIGNAL_BUFFER (1024 * 1024 * 4)

#define DEBUG_SIGNALS 0

#define SIGNAL_SHOW_PROPERTIES 0
#define DEBUG_CONNECT 1

struct Signal
{
  unsigned byUNL : 1;
  unsigned generator : 1;
  union 
  {
    char *UNL;
    NodeType *node;
  } source;
  Socket *socket;

  TypeID type;
  //int type; // type of value
  unsigned int eltsize;
  TypeID convtype;
  unsigned int conveltsize; // bytes/value in I/O buffers, if different
  unsigned int columns; // values/element
  unsigned int rows; // elements/packet
  unsigned long hz; // number of elements in a second

  unsigned int length; // values/packet
  Identity *identity;
  char *receiverAuth;
  struct timeval receipt;
  char *storage;
  size_t stored;
  size_t allocated;

  // The label for the first dimension
  char *label;
  char **indexLabels;
};

int GetSignalProxySocket(Socket **sockPtr, char *UNL);

static void setLength(Signal *signal)
{
  signal->length = signal->rows * signal->columns;
}

static Signal *allocSignal()
{
  Signal *ret = malloc(sizeof(Signal));
  ret->socket = NULL;
  ret->identity = NULL;
  ret->conveltsize = 0;
  ret->eltsize = 0;
  ret->receiverAuth = NULL;
  ret->label = NULL;
  ret->indexLabels = NULL;
  ret->rows = 1;
  ret->columns = 0;
  ret->length = 0;
  ret->hz = 0;
  ret->type = T_EMPTY;
  return ret;
}

Signal *NewSignalGenNode()
{
  Signal *ret = allocSignal();
  ret->byUNL = 0;
  ret->socket = MultipointListen(0, 1);
  ret->source.node = CreateAddressNode("", T_SIGNAL, 
				       SocketLocal(ret->socket));
  ret->generator = 1;
  return ret;
}

Signal *NewSignalGenUNL(char *UNL)
{
  Signal *ret = allocSignal();
  NodeType *node;
  LockUNLInfo lockInfo;
  ret->byUNL = 1;
  ret->socket = MultipointListen(0, 1);
  ret->source.UNL = UNL;
  node = CreateAddressNode("", T_SIGNAL, SocketLocal(ret->socket));

  LockUNL(NULL, &lockInfo);
  lockInfo.common.operationFlags = ENCHANT_LOCK_REMOVE;
  if (!LockUNL(UNL, &lockInfo))
    {
      printf("Couldn't lock '%s'\n", UNL);
    }
  else if (!PublishNode(UNL, node, NULL))
    {
      printf("Couldn't publish to '%s'\n", UNL);
    }
  FreeNode(node);
  ret->generator = 1;
  return ret;
}

Signal *NewSignalGenNone()
{
  Signal *ret = allocSignal();
  ret->byUNL = 1;
  ret->socket = MultipointListen(0, 1);
  ret->source.UNL = NULL;
  ret->generator = 1;
  return ret;
}

void SignalGenGetReceipts(Signal *signal, char *auth)
{
  signal->receiverAuth = strdup(auth);
  signal->storage = NULL;
  signal->stored = 0;
  signal->allocated = 0;
  gettimeofday(&signal->receipt, NULL);
}

static int negotiateSend(Socket *setup, Signal *sig)
{
  int done = 0;
  int index;
  int use_ssl = 0;
  int sent_shape = 0;
  IntroMessage message;

  if (DEBUG_SIGNALS)
    printf("Sending properties\n");

  do
    {
      done = 1;
      if (!sent_shape)
	done = 0;
      do
	{
	  index = IntroRecv(setup, receiverProperties, &message);
	  if (index <= 0)
	    ;
	  else if (message.format == &protocolEncrypt)
	    {
	      if (!strcmp(message.value.string_val, PROTOCOL_ENCRYPT_SSL))
		{
		  printf("Will use SSL\n");
		  use_ssl = 1;
		}
	    }
	  if (index)
	    done = 0;
	}
      while (index > 0);
      if (use_ssl)
	{
	  if (DEBUG_SIGNALS)
	    printf("Switching to secure connection.\n");
	  SocketSecure(setup, sig->identity, 1);
	}
      if (!done)
	{
	  done = 1;
	  if (!sent_shape)
	    {
	      message.format = &signalType;
	      message.value.int_val = 
		(sig->type & T_NUMERIC_FORMAT_MASK) >> 4;
	      IntroSend(setup, &message);
	      
	      message.format = &signalEltsize;
	      message.value.int_val = sig->eltsize;
	      IntroSend(setup, &message);
	      
	      message.format = &signalLength;
	      message.index = 1;
	      message.value.int_val = sig->columns;
	      IntroSend(setup, &message);
	      
	      message.format = &signalLength;
	      message.index = 2;
	      message.value.int_val = sig->rows;
	      IntroSend(setup, &message);
	      
	      if (sig->label)
		{
		  message.format = &signalLabel;
		  message.index = 1;
		  message.value.string_val = sig->label;
		  IntroSend(setup, &message);
		}

	      if (sig->indexLabels)
		{
		  int i;
		  message.format = &signalIndexLabel;
		  for (i = 0; i < sig->columns; i++)
		    {
		      message.index = i + 1;
		      message.value.string_val = sig->indexLabels[i];
		      IntroSend(setup, &message);
		    }
		}
	      
	      if (sig->hz)
		{
		  message.format = &signalRate;
		  message.value.int_val = sig->hz;
		  IntroSend(setup, &message);
		}
	      
	      if (DefaultIdentity() || sig->identity)
		{
		  if (DEBUG_SIGNALS)
		    printf("Sending SSL prop\n");
		  
		  message.format = &protocolEncrypt;
		  message.value.string_val = PROTOCOL_ENCRYPT_SSL;
		  IntroSend(setup, &message);
		}
	      sent_shape = 1;
	      done = 0;
	    }
	  IntroDone(setup);
	}

      if (DEBUG_SIGNALS)
	printf("Done sending properties\n");
    }
  while (!done);
  if (DEBUG_SIGNALS)
    printf("Done negotiating\n");
  return 1;
}

Signal *NewSignalNode(NodeType *node)
{
  Signal *ret = allocSignal();
  ret->byUNL = 0;
  ret->source.node = node;
  ret->generator = 0;
  return ret;
}

Signal *NewSignalUNL(char *UNL)
{
  Signal *ret = allocSignal();
  ret->byUNL = 1;
  ret->source.UNL = UNL;
  ret->generator = 0;
  return ret;
}

static int negotiateRecv(Signal *signal)
{
  int done = 1;
  int use_ssl = 0;
  int using_ssl = 0;
  int sent_unl = 0;
  int sent_interaction = 0;
  IntroMessage message;
  signal->type = 0;
  do
    {
      int index;
      done = 1;
#ifndef WANT_SSL
      // Otherwise we'll accept it...
      use_ssl = 0;
#endif
      if (!sent_interaction)
	{
	  message.format = &protocolInteraction;
	  message.value.string_val = PROTOCOL_TYPE_SIGNAL;
	  IntroSend(signal->socket, &message);
	  done = 0;
	  sent_interaction = 1;
	}
      else if (!sent_unl)
	{
	  sent_unl = 1;
	  message.format = &signalUNL;
	  message.value.string_val = signal->source.UNL;
	  IntroSend(signal->socket, &message);
	  done = 0;
	}
      if (use_ssl && !using_ssl)
	{
	  if (DEBUG_SIGNALS)
	    printf("Confirming encryption\n");
	  done = 0;
	  message.format = &protocolEncrypt;
	  message.value.string_val = PROTOCOL_ENCRYPT_SSL;
	  IntroSend(signal->socket, &message);
	  if (DEBUG_SIGNALS)
	    printf("Done confirming encryption\n");
	}
      IntroDone(signal->socket);
      // Any others here...
      if (use_ssl && !using_ssl)
	{
	  SocketSecure(signal->socket, signal->identity, 0);
	  using_ssl = 1;
	}
      if (!done)
	{
	  done = 1;
	  do
	    {
	      index = IntroRecv(signal->socket, generatorProperties, &message);
	      if (index <= 0)
		;
	      else if (message.format == &signalType)
		{
		  signal->type = (message.value.int_val << 4) + T_NUMERIC;
		  if (SIGNAL_SHOW_PROPERTIES)
		    printf("type: 0x%x\n", signal->type);
		}
	      else if (message.format == &signalEltsize)
		{
		  signal->eltsize = message.value.int_val;
		  if (SIGNAL_SHOW_PROPERTIES)
		    printf("eltsize: %d\n", signal->eltsize);
		}
	      else if (message.format == &signalLength)
		{
		  if (message.index == 2)
		    {
		      signal->rows = message.value.int_val;
		      if (SIGNAL_SHOW_PROPERTIES)
			printf("rows: %d\n", signal->rows);
		    }
		  if (message.index == 1)
		    {
		      signal->columns = message.value.int_val;
		      if (SIGNAL_SHOW_PROPERTIES)
			printf("columns: %d\n", signal->columns);
		    }
		}
	      else if (message.format == &signalLabel)
		{
		  if (SIGNAL_SHOW_PROPERTIES)
		    printf("label: %s\n", message.value.string_val);
		  signal->label = message.value.string_val;
		}
	      else if (message.format == &signalIndexLabel)
		{
		  if (SIGNAL_SHOW_PROPERTIES)
		    printf("index label %d: %s\n", 
			   message.index,
			   message.value.string_val);
		}
	      else if (message.format == &signalRate)
		{
		  if (SIGNAL_SHOW_PROPERTIES)
		    printf("Sample rate: %d\n", message.value.int_val);
		  signal->hz = message.value.int_val;
		}
	      else if (message.format == &protocolEncrypt)
		{
		  if (SIGNAL_SHOW_PROPERTIES)
		    printf("encrypt: %s\n", message.value.string_val);
		  if (!strcmp(message.value.string_val, PROTOCOL_ENCRYPT_SSL))
		    use_ssl = 1;
		}
	      if (index)
		done = 0;
	    }
	  while (index > 0);
	}
      if (DEBUG_SIGNALS)
	printf("Done reading properties\n");
    }
  while (!done);
  setLength(signal);
  if (!signal->type)
    {
      if (DEBUG_SIGNALS)
	printf("Signal interaction refused.\n");
      return 0;
    }
  /*
  printf("signal: %d values of %d bytes each: %d bytes/packet\n",
	 signal->length, signal->eltsize, signal->size);
  */
  return 1;
}

static int doConnect(Signal *signal, NodeType *target)
{
  signal->socket = SocketConnect(target->data.netsock_val, 1);
  if (signal->socket)
    return negotiateRecv(signal);
  return 0;
}

int SignalConnect(Signal *signal)
{
  NodeType *target;
  char *unl;
  if (signal->socket)
    return 1;
  //printf("Connecting\n");
  if (!signal->byUNL)
    return doConnect(signal, signal->source.node);
  unl = signal->source.UNL;
  target = GetNode(unl, NULL);
  if (target && doConnect(signal, target))
    return 1;
  return 0;
}

int SignalConnectWait(Signal *signal)
{
  NodeType *target;
  char *unl;
  SubscribeUNLInfo subInfo;
  if (signal->socket)
    return 1;
  //printf("Connecting\n");
  if (!signal->byUNL)
    return doConnect(signal, signal->source.node);
  unl = signal->source.UNL;
  SubscribeUNL(NULL, &subInfo);
  subInfo.common.operationFlags = ENCHANT_SUBSCRIBE_INITIAL;
  if (!SubscribeUNL(unl, &subInfo))
    {
      printf("Couldn't wait\n");
      return 0;
    }
  unl = subInfo.common.actualUNL;
  do
    {
      do
	{
	  if (DEBUG_CONNECT)
	    printf("Now scanning %s\n", unl);
	  target = WaitForUpdate(NULL, unl);
	  if (DEBUG_CONNECT)
	    printf("Got update for %s\n", unl);
	  // If target is NULL, we've picked up the unpublish, which we
	  // will just ignore.
	}
      while (!target);
      if (doConnect(signal, target))
	return 1;
      printf("Could not connect to signal directly\n");
      if (GetSignalProxySocket(&signal->socket, signal->source.UNL))
	{
	  if (negotiateRecv(signal))
	    return 1;
	}
      printf("Nor by proxy\n");
    }
  while (1);
}

int SignalConnectByProxy(Signal *signal)
{
  // Only works if we have the signal location as a UNL, since the server
  // name is needed to be the proxy
  if (!signal->byUNL)
    return 0;
  if (!GetSignalProxySocket(&signal->socket, signal->source.UNL))
    return 0;
  return negotiateRecv(signal);
}

int SignalGenConnected(Signal *signal, Socket *socket)
{
  if (!negotiateSend(socket, signal))
    return 0;
  SocketJoinMultipoint(socket, signal->socket);
  SocketClose(socket);
  return 1;
}

int ScanSignalElements(FILE *file, Signal *signal,
		       struct timeval *time, char *buffer)
{
  char str[9];
  int sz;
  int rows;
  int cols;
  int i;
  int ret;
  TypeID type = 0;
  int conveltsize;
  ret = fscanf(file, "%ld.%ld%8c%d %dx%d [ ", &time->tv_sec, &time->tv_usec,
	       str, &sz, &rows, &cols);
  if (ret != 6)
    {
      printf("Failed to read line\n");
      return 0;
    }
  str[8] = '\0';
  if (!strcmp("     INT", str))
    type = T_INT_FORMAT;
  else if (!strcmp("   FLOAT", str))
    type = T_FLOAT_FORMAT;
  else
    printf("Type was '%s'\n", str);
  if (!buffer)
    {
      signal->rows = rows;
      signal->columns = cols;
      signal->type = type;
      signal->eltsize = sz;
      setLength(signal);
    }
  conveltsize = signal->conveltsize ? signal->conveltsize : signal->eltsize;
  if (rows != signal->rows || cols != signal->columns || 
      type != signal->type || sz != signal->eltsize)
    {
      printf("Rows: %d should be %d; cols: %d should be %d;\n"
	     "type: %x should be %x; eltsize: %d should be %d\n", rows,
	     signal->rows, cols, signal->columns, type, signal->type,
	     sz, signal->eltsize);
      return 0;
    }
  for (i = 0; i < signal->length; i++)
    {
      if (type == T_INT_FORMAT)
	{
	  long value;
	  ret = fscanf(file, "%ld", &value);
	  if (!buffer)
	    continue;
	  if (conveltsize == 4)
	    ((long *)buffer)[i] = value;
	  else if (conveltsize == 2)
	    ((short *)buffer)[i] = value;
	  else
	    ((signed char *)buffer)[i] = value;
	}
      else if (type == T_FLOAT_FORMAT)
	{
	  double value;
	  ret = fscanf(file, "%lg", &value);
	  if (!buffer)
	    continue;
	  if (conveltsize == 8)
	    ((double *)buffer)[i] = value;
	  else if (conveltsize == 4)
	    ((float *)buffer)[i] = value;
	}
    }
  ret = fscanf(file, " %1c\n", str);
  return 1;
}

static char *types[] =
{
  "INT",
  "FLOAT",
  "FIXED",
  "FRACTION"
};

static char *SignalTypeName(int type)
{
  if (type < (sizeof(types) / sizeof(char *)))
    return types[type];
  else
    return "UNKNOWN";
}

void PrintSignalElements(FILE *file, Signal *signal, 
			 struct timeval time, char *buffer)
{
  int i;
  int conveltsize;
  fprintf(file, "%06ld.%06ld %7s%d %dx%d [ ", time.tv_sec, time.tv_usec, 
	 SignalTypeName((signal->type >> 8) & 0xFF), signal->eltsize,
	 signal->rows, signal->columns);
  if (signal->conveltsize)
    conveltsize = signal->conveltsize;
  else
    conveltsize = signal->eltsize;
  for (i = 0; i < signal->length; i++)
    {
      if ((signal->type & T_NUMERIC_FORMAT_MASK) == T_FLOAT_FORMAT)
	{
	  if (conveltsize == 8)
	    {
	      fprintf(file, "%g ", ((double *)buffer)[i]);
	    }
	  else if (conveltsize == 4)
	    {
	      fprintf(file, "%g ", ((float *)buffer)[i]);
	    }
	}
      else
	{
	  long val;
	  if (conveltsize == 1)
	    val = ((signed char *)buffer)[i];
	  else if (conveltsize == 2)
	    val = ((signed short *)buffer)[i];
	  else if (conveltsize == 4)
	    val = ((long *)buffer)[i];
	  if ((signal->type & T_NUMERIC_FORMAT_MASK) == T_INT_FORMAT)
	    fprintf(file, "%ld ", val);
	  else if ((signal->type & T_NUMERIC_FORMAT_MASK) == T_FIXED_FORMAT)
	    fprintf(file, "%g ", val/((double) ((long long) 1 << (conveltsize << 2))));
	  else if ((signal->type & T_NUMERIC_FORMAT_MASK) == T_FRACTION_FORMAT)
	    fprintf(file, "%g ", val/((double) ((long long) 1 << (conveltsize << 3))));
	}
    }
  fprintf(file, "]\n");
}

static void handleAck(Signal *sig, struct timeval time)
{
  int posn = 0;
  struct timeval storedTime;
  do
    {
      storedTime = *((struct timeval *)(sig->storage + posn));
      if (storedTime.tv_sec > time.tv_sec ||
	  (storedTime.tv_sec == time.tv_sec &&
	   storedTime.tv_usec > time.tv_usec))
	return;
      posn += sig->length * sig->eltsize + sizeof(struct timeval);
    }
  while (storedTime.tv_usec != time.tv_usec || 
	 storedTime.tv_sec != time.tv_sec);
  memmove(sig->storage, sig->storage + posn, sig->stored - posn);
  sig->stored -= posn;
}

static void sendMissedSamples(Signal *sig, Socket *socket)
{
  int posn = 0;
  int i = 0;
  unsigned long item;
  struct timeval time;
  if (!sig->receiverAuth)
    return;
  while (posn < sig->stored)
    {
      time = *((struct timeval *)(sig->storage + posn));
      SocketSendStart(socket);

      item = htonl(time.tv_sec);
      SocketSend(socket, (char *) &item, 4);

      // timestampe microseconds
      item = htonl(time.tv_usec);
      SocketSend(socket, (char *) &item, 4);

      SocketSend(socket, sig->storage + posn + sizeof(struct timeval), 
		 sig->length * sig->eltsize);

      SocketSendEnd(socket);

      posn += sig->length * sig->eltsize + sizeof(struct timeval);
      i++;
    }
}

void SignalGenAccept(Signal *sig)
{
  Socket *setup;
  setLength(sig);
  while ((setup = MultipointAccept(sig->socket)))
    {
      negotiateSend(setup, sig);
      sendMissedSamples(sig, setup);
      SocketClose(setup);
    }
}

int SignalSend(Signal *sig, struct timeval time, void *buffer)
{
  unsigned long item;

  if (!sig->generator)
    return 0;

  ntoh(buffer, sig->conveltsize ? sig->conveltsize : sig->eltsize,
       sig->eltsize, sig->length, sig->type);

  SocketSendStart(sig->socket);

  // timestamp seconds
  item = htonl(time.tv_sec);
  SocketSend(sig->socket, (char*) &item, 4);

  // timestampe microseconds
  item = htonl(time.tv_usec);
  SocketSend(sig->socket, (char*) &item, 4);

  // send entire vector contents in one operation.
  SocketSend(sig->socket, (char *)buffer, sig->eltsize * sig->length);

  SocketSendEnd(sig->socket);

  if (DEBUG_SIGNALS)
    printf("Sent packet\n");

  if (sig->receiverAuth)
    {
      // Check for receipts
      int fd;
      int ret;
      char temp[10];
      struct timeval resp;
      do 
	{
	  ret = SocketRecvMulti(sig->socket, temp, 10, &fd);
	  if (ret == 10 && temp[0] == 'r' && temp[1] == 8)
	    {
	      resp.tv_sec = ntohl(((long *)(temp + 2))[0]);
	      resp.tv_usec = ntohl(((long *)(temp + 2))[1]);
	      handleAck(sig, resp);
	    }
	}
      while (ret == 10);

      // Store this sample
      if (!sig->storage)
	{
	  sig->allocated = (sig->eltsize * sig->length + 
			    sizeof(struct timeval)) * 16;
	  sig->storage = malloc(sig->allocated);
	}
      else if (sig->allocated < sig->stored + sig->eltsize * sig->length + 
	       sizeof(struct timeval))
	{
	  if (sig->allocated * 2 > MAX_SIGNAL_BUFFER)
	    return 0;
	  sig->allocated *= 2;
	  sig->storage = realloc(sig->storage, sig->allocated);
	}
      memcpy(sig->storage + sig->stored, &time, sizeof(struct timeval));
      sig->stored += sizeof(struct timeval);
      memcpy(sig->storage + sig->stored, buffer, sig->eltsize * sig->length);
      sig->stored += sig->eltsize * sig->length;
    }

  return 1; // success
}

static void closeConnection(Signal *sig)
{
  SocketClose(sig->socket);
  sig->socket = NULL;
}

int SignalRecv(Signal *sig, struct timeval *time_p, void *buffer)
{
  size_t numBytes;
  unsigned long item;

  if (sig->generator)
    {
      int posn;
      if (!sig->stored)
	return 0;
      *time_p = *((struct timeval *)(sig->storage));
      
      memcpy(buffer, sig->storage + sizeof(struct timeval),
	     sig->length * sig->eltsize);
      
      posn = sig->length * sig->eltsize + sizeof(struct timeval);
      memmove(sig->storage, sig->storage + posn, sig->stored - posn);
      sig->stored -= posn;
      return 1;
    }

  if (!sig->socket) 
    {
      SignalConnectWait(sig);
      if (!sig->socket) 
	{
	  fprintf(stderr, "unable to open socket\n");
	  return 0;
	}
    }

  // get timestamp seconds
  numBytes = SocketRecv(sig->socket, (char *)&item, 4);
  if (numBytes != 4) 
    {
      closeConnection(sig);
      return 0;
    }
  time_p->tv_sec = ntohl(item);
  
  // get timestamp microseconds
  numBytes = SocketRecv(sig->socket, (char *)&item, 4);
  if (numBytes != 4) 
    {
      fprintf(stderr, "failed to get timestamp microseconds\n");
      closeConnection(sig);
      return 0;
    }
  time_p->tv_usec = ntohl(item);
  
  // get vector contents
  numBytes = SocketRecv(sig->socket, buffer, sig->eltsize * sig->length);
  if (numBytes != sig->eltsize * sig->length)
    {
      fprintf(stderr, "error: receved only %d of %d bytes\n", 
	      numBytes, sig->eltsize * sig->length);
      closeConnection(sig);
      return 0;
    }
  // convert byteorder
  ntoh(buffer, sig->eltsize, 
       sig->conveltsize ? sig->conveltsize : sig->eltsize, 
       sig->length, sig->type);

  if (DEBUG_SIGNALS)
    printf("Got packet\n");

  return 1;
}

void SignalRecvd(Signal *sig, struct timeval timestamp)
{
  long item;
  char ch;
  SocketSendStart(sig->socket);

  ch = 'r';
  SocketSend(sig->socket, &ch, 1);

  ch = 8;
  SocketSend(sig->socket, &ch, 1);

  // timestamp seconds
  item = htonl(timestamp.tv_sec);
  SocketSend(sig->socket, (char*) &item, 4);

  // timestampe microseconds
  item = htonl(timestamp.tv_usec);
  SocketSend(sig->socket, (char*) &item, 4);

  SocketSendEnd(sig->socket);
}

void SignalClose(Signal *signal)
{
  if (signal->byUNL && signal->generator)
    {
      // We do this first, so that people don't immediately get us again...
      PublishNode(signal->source.UNL, NULL, NULL);
    }
  if (signal->socket)
    closeConnection(signal);
  if (signal->label)
    free(signal->label);
  if (signal->receiverAuth)
    free(signal->receiverAuth);
  free(signal);
}

void SignalCopyProperties(Signal *source, Signal *target)
{
  target->type = source->type;
  target->eltsize = source->eltsize;
  target->rows = source->rows;
  target->columns = source->columns;
  setLength(target);
}

void SignalSetSampleRate(Signal *signal, unsigned hz, unsigned length)
{
  if (!signal->generator)
    {
      fprintf(stderr, "Changing the sample rate for a received signal is "
	     "not currently supported\n");
      return;
    }
  signal->hz = hz;
  signal->rows = length;
  setLength(signal);
}

void SignalSetDimension(Signal *signal, int dimension, 
			unsigned length, char *label, char **indexLabels)
{
  if (!signal->generator)
    {
      fprintf(stderr, "Changing the signal dimensions for a received signal "
	      "is not supported.\n");
      return;
    }
  if (dimension == 1)
    {
      if (label)
	signal->label = strdup(label);
      else
	signal->label = NULL;
      if (indexLabels)
	{
	  signal->indexLabels = indexLabels;
	}
      else
	{
	  signal->indexLabels = NULL;
	}
      signal->columns = length;
      setLength(signal);
    }
  else
    {
      fprintf(stderr, "Dimensions other than 1 not supported yet.\n");
    }
}

static void SignalSetSizeType(Signal *signal, size_t eltsize, TypeID type)
{
  if (signal->generator)
    {
      signal->type = type;
      signal->eltsize = eltsize;
    }
  else
    {
      if ((type & T_NUMERIC_FORMAT_MASK) != 
	  (signal->type & T_NUMERIC_FORMAT_MASK))
	{
	  fprintf(stderr, "Changing the element type of received signals "
		  "from %x to %x not currently supported.\n",
		  signal->type, type);
	}
      signal->conveltsize = eltsize;
    }
}

void SignalSetType(Signal *signal, TypeID type)
{
  SignalSetSizeType(signal, 1 << (type & 0xF), type);
}

TypeID SignalGetType(Signal *signal)
{
  return ffs(signal->eltsize) + signal->type;
}

int SignalClientCount(Signal *signal)
{
  return SocketEndpoints(signal->socket);
}

NetSocketLocation *SignalSource(Signal *signal)
{
  if (signal->generator)
    return NULL;
  return SocketRemote(signal->socket);
}

int SignalConnected(Signal *signal)
{
  return !!signal->socket;
}

// uses length
size_t SignalBufferSize(Signal *signal)
{
  if (signal->conveltsize > signal->eltsize)
    return signal->length * signal->conveltsize;
  else
    return signal->length * signal->eltsize;
}

// XXXX eliminate this when possible

Socket *IntSignalGetSocket(Signal *signal)
{
  return signal->socket;
}

void SignalSetIdentity(Signal *signal, Identity *identity)
{
  signal->identity = identity;
}

size_t SignalGetEltsize(Signal *signal)
{
  return signal->eltsize;
}

int SignalGetLength(Signal *signal, int dimension)
{
  if (dimension == 1)
    return signal->columns;
  return 0;
}

int SignalGetUnitSamples(Signal *signal)
{
  return signal->rows;
}

unsigned SignalGetSampleRate(Signal *signal)
{
  return signal->hz;
}
