/*************************************************
 * File: socket.h
 * Author: Daniel Barkalow
 * Birth Date: 2/4/03
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2003 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *************************************************/

#ifndef _SOCKET_H
#define _SOCKET_H

#include <enchantment/types.h>
#include <enchantment/identity.h>
#include <unistd.h>
#include <sys/types.h>

typedef struct Socket Socket;
typedef struct SocketFactory SocketFactory;

/*
 * Sockets are a general mechanism for the transfer of records, which
 * are chunks of binary data such that each record is intended to be
 * read separately.
 *
 * A file socket reads from or writes to a local file.
 *
 * A multipoint socket broadcasts records to all connected sockets; it
 * is created directly, and a function is used to update the list of
 * connected clients, but the clients are not normally treated
 * individually. A count of connected clients is available so that it
 * is possible to know whether it is worth producing data. 
 *
 * A program which wants to be connected to by other programs which it
 * will treat individually listens to get a SocketFactory. It then
 * gets sockets by accepting connections.
 *
 * A program can connect to remote programs to get a socket; it does
 * not need to consider whether the other end is single point or
 * multipoint. It can request a reliable socket or an unreliable one.
 *
 * Note that this code is no good for connections involving other
 * code, because it uses its own protocol for exchanging the records
 * and for negotiating the start of the connections.
 */

/* Returns a socket connecting to a remote target. If reliable is
 * false, not all of the records send by the other end will
 * necessarily arrive and records may be returned out of order,
 * although SocketRecv will still only return complete records. The
 * other end may not allow unreliable connections, in which case the
 * connection will be reliable regardless of this flag.
 */
Socket *SocketConnect(NetSocketLocation *target, int reliable);

SocketFactory *SocketListen(int port, int allowUnreliable);
Socket *SocketAccept(SocketFactory *factory);

/* Multipoint has a main socket, which broadcasts to all connected
 * clients. When clients are accepted, there are also sockets used to
 * send initial information. Once the initialization socket is closed,
 * the client will get information from the main socket. */
Socket *MultipointListen(int port, int allowUnreliable);
Socket *MultipointAccept(Socket *socket);

/* Makes the given socket an initialization socket for the given multipoint
 * socket. */
int SocketJoinMultipoint(Socket *socket, Socket *multi);

/* Returns the number of other ends of this socket. Valid for all
 * types, although only interesting for multipoint sockets.
 */
unsigned SocketEndpoints(Socket *socket);

/* Returns the endpoints of the socket. The remote end is NULL for a
 * multipoint socket or a file. The local end is NULL for a file and
 * may be incomplete for a multipoint socket (see NetSocketLocationMerge).
 */
NetSocketLocation *SocketLocal(Socket *socket);
NetSocketLocation *SocketRemote(Socket *socket);

/* Modifies target so that it will target the host specified by host.
 * This is useful when the target was returned by SocketLocal on a
 * multipoint socket, which may not be able to determine the correct
 * host information; a remote machine with a connection to this host
 * may have a better idea.
 */
void NetSocketLocationMerge(NetSocketLocation *target, 
			    NetSocketLocation *host);

/* Returns a socket which connects to a local file given a file
 * descriptor for it.
 */
Socket *SocketFile(int fd);

/* Sends a record of the given size to the socket. Caller is
 * responsible for the memory. 
 */
int SocketSendRecord(Socket *socket, char *buffer, size_t size);

/* SocketSend and SocketSendRecord calls between a SocketSendStart and
 * a SocketSendEnd on the socket will be combined and either all will arrive
 * or none will arrive.
 */
void SocketSendStart(Socket *socket);
int SocketSendEnd(Socket *socket);

int SocketSend(Socket *socket, char *buffer, size_t size);

int SocketRecv(Socket *socket, char *buffer, size_t size);

/* Reads from one of the connections in this multipoint socket that
 * has sent data, and sets *client to the fd of the connection. Returns -1
 * if no client has sent any data that has not been read.
 */
int SocketRecvMulti(Socket *socket, char *buffer, size_t size, int *client);

/* These functions allow sending a large data object such that the receiving
 * end can determine the end of the object without external information.
 *
 * On the writing side, call BeginObject, Send multiple times,
 * and then EndObject. On the reading side, call BeginObject, Read until it
 * returns 0 size, and then EndObject.
 */
int SocketBeginObject(Socket *socket);
int SocketEndObject(Socket *socket);

/* Returns a record from the socket. The memory belongs to the socket
 * and only lasts until the next socket call on the same socket.
 */
int SocketRecvRecord(Socket *socket, char **buffer, size_t *size);

/* Closes the socket, freeing its memory and releasing all of its resources. */
void SocketClose(Socket *socket);

/* Closes the socket factory, freeing its memory and releasing all of its
 * resources. Sockets derived from the factory may still be used. */
void SocketFactoryClose(SocketFactory *factory);

/* Waits for activity on one or more sockets from the array or (if provided),
 * activity on the socket factory. fd_set holds the file descriptors for the
 * sockets that had activity. */
void SocketSelect(fd_set *fds, DArray socketP_Array, SocketFactory *factory);

/* Tests whether the given socket signalled activity in the given fd_set.
 */
int SocketIncluded(fd_set *fds, Socket *sock);

/* Adds the file descriptor(s) for the reading portion of the given
 * socket to the given set, and returns the maximum file descriptor
 * set by the call. Returns -1 if there are no appropriate file
 * descriptors. Since this call presumably leads to attempting to read
 * from the socket, it is flushed if necessary.
 */
int SocketInclude(fd_set *fds, Socket *sock);

/* Returns a distintive name for a socket. */
char *SocketName(Socket *socket);

/* Returns true if a read from sock would not (immediately) block; waits
* up to waitUsec. */
int SocketReadReady(Socket *sock, unsigned waitUsec);

/* Flushes writes to the socket. */
int SocketFlush(Socket *socket);

/* Makes the socket secure. Both sides must call this at the same time by
 * some prior arrangement. If identity is NULL, the default identity is used.
 * If there is no provided identity, not default identity, and server is set,
 * the call will certainly fail.
 *
 * Returns true if encryption could be negotiated. Note that both
 * sides must call this at the same time, even if it does not work.
 *
 * BUG: if WANT_SSL is not defined, and we call this at the same time
 * as the other end calls it with WANT_SSL defined, we get messed
 * up. We should support doing a (failing) handshake in our code if
 * WANT_SSL is not defined.
 */
int SocketSecure(Socket *socket, Identity *identity, int server);

/* Tests whether a socket is secure. */
int SocketIsSecure(Socket *socket);

/* Returns the fingerprint of the remote end of the socket, if there is a
 * certificate for it.
 */
char *SocketRemoteFingerprint(Socket *socket);

/* Returns the error from the most recent socket call on this socket. */
int SocketError(Socket *socket);

void PrintLocation(NetSocketLocation *location);

#endif
