/*************************************************
 * File: tree.c
 * Author: Jonathan Gips
 * Birth Date: 6/27/01
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * 
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *************************************************/

#include <enchantment/node.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include <string.h>
#include <assert.h>
#include <enchantment/darray.h>
#include <enchantment/dstring.h>
#include <netinet/in.h>
#include <enchantment/unl.h>

#include <sys/time.h>

#define IsBranch(x) (((x) & FORMAT_MASK) == T_BRANCH)

/**** Local node construction functions ****/

static int PointerCompFn(const void *elemptr1, const void *elemptr2)
{
  void *el1 = *((void**)elemptr1), *el2 = *((void**)elemptr2);
  
  if (el1 == el2)
    return 0;
  else if ((uintptr_t) el1 < (uintptr_t) el2)
    return -1;
  else
    return 1;
}

static int NodeNameCompFn(const void *elem1, const void *elem2)
{
  NodeType *node1 = *((NodeType**)elem1), *node2 = *((NodeType**)elem2);
  
  if (node1->name == node2->name) 
    return 0;
  else if ((uintptr_t) node1 < (uintptr_t) node2) 
    return -1;
  else
    return 1;
}

static void NodeFreeFn(void *elem)
{
  FreeNode(*((NodeType **) elem));
}

NodeType *CreateTypedNode(const char *name, TypeID type, DataType *data)
{
  NodeType *newNode;
  
  // Create basic node 
  newNode = (NodeType *) malloc(sizeof(NodeType));
 
  newNode->name = InstanceString(name);
  newNode->type = type;
  newNode->local = NULL;
  
  // Fill in the data field with a NULL entry if pointer not passed in
  if (!data)
    {
      newNode->data.empty_val = NULL;
    }
  else
    {
      newNode->data = *data; // DataType union is stored within the node 
    }


  /**** Internal Tree field initialization ****/
  if (IsBranch(newNode->type)) 
    {
      newNode->data.branch_val = ArrayNew(sizeof(NodeType *), 1, NodeFreeFn);
    }

  return newNode;
}

double NodeGetDouble(NodeType *node)
{
  switch (node->type)
    {
    case T_DOUBLE:
      return *node->data.doubleptr_val;
    case T_FLOAT:
      return node->data.float_val;
    case T_INT32:
    case T_INT16:
    case T_INT8:
      return node->data.int_val;
    default:
      return 0;
    }
}

void FreeNode(NodeType *node)
{  
  if (node == NULL) 
    return;
  
  switch (GetFormat(node->type))
    {
    case T_STRING:
    case T_SYMLINK:
    case T_DOUBLE:
    case T_NETSOCK:
    case T_TIMEVAL:
      if (node->data.empty_val) 
	{
	  free(node->data.empty_val);
	}
      break;
    case T_BRANCH:
      if (node->data.branch_val)
	{
	  ArrayFree(node->data.branch_val);
	}
      break;
    }
  node->data.empty_val = NULL;
  
  // Free this node's structure
  free(node);
}

NodeType *CreateBranchNode(const char *name, TypeID type)
{
  NodeType *newNode;
 
  if (!IsBranch(type))
    {
      printf("ERROR: trying to create a non-branch with CreateBranchNode\n");
      return NULL; // enforce correct types
    }
  
  // Branches always own their data
  newNode = CreateTypedNode(name, type, NULL);
  
  return newNode;
}

NodeType *CreateAddressNode(const char *name, TypeID type, 
			    NetSocketLocation *location)
{
  NodeType *newNode;
  DataType *holder = malloc(sizeof(DataType));
  holder->netsock_val = CopyNetSocketLocation(location);
  newNode = CreateTypedNode(name, type, holder);
  return newNode;
}

NodeType *CreateIntegerNode(const char *name, TypeID type,
			    long value)
{
  NodeType *ret = CreateTypedNode(name, type, NULL);
  ret->data.int_val = value;
  return ret;
}

NodeType *CreateStringNode(const char *name, TypeID type, char *value)
{
  NodeType *ret = CreateTypedNode(name, type, NULL);
  ret->data.string_val = strdup(value);
  return ret;
}

NodeType *CreateEmptyNode(const char *name, TypeID type)
{
  return CreateTypedNode(name, type, NULL);
}

void NodeSetString(NodeType *node, char *value)
{
  free(node->data.string_val);
  node->data.string_val = strdup(value);
}

NodeType *CreateTimeNode(const char *name, TypeID type, struct timeval *value)
{
  NodeType *ret = CreateTypedNode(name, type, NULL);
  ret->data.timeval_val = malloc(sizeof(struct timeval));
  if (value)
    *ret->data.timeval_val = *value;
  else
    gettimeofday(ret->data.timeval_val, NULL);
  return ret;
}

static int ContainsNode(NodeType *parentNode, NodeType *descendNode)
{
  if (parentNode == descendNode) 
    return 1;
  else if (IsBranch(parentNode->type))
    {
      unsigned index;
      NodeType *child = NULL;
      
      for (index = 0; (child = BranchIndex(parentNode, index)); index++)
	{
	  if (ContainsNode(child, descendNode)) 
	    return 1;
	}
      
      return 0;
    }
  else
    return 0;
}

int InsertNode(NodeType *intoNode, NodeType *node)
{
  int index;
  NodeType *existNode = NULL;
  
  if (!intoNode || !node) 
    {
      printf("InsertNode failed: NULL pointer\n");
      return 0;
    }

  if (!IsBranch(intoNode->type))
    {
      printf("InsertNode failed: not into branch\n");
      return 0;
    }
      
  // Guard against loops here.
  if (ContainsNode(node, intoNode)) 
    {
      printf("InsertNode failed: loop\n");
      return 0;
    }
      
  // Search for index of a node with the same name
  index = ArraySearch(intoNode->data.branch_val, &node, 
		      NodeNameCompFn, 0, 0);
  
  if (index == -1) 
    {
      ArrayAppend(intoNode->data.branch_val, &node);
    }
  else
    {
      existNode = *(NodeType **)ArrayNth(intoNode->data.branch_val, index);
      if (existNode == node) 
	return 1;
      ArrayReplaceAt(intoNode->data.branch_val, &node, index);
    }
  
  return 1;
}

NodeType *GetChild(NodeType *parentNode, char *childName)
{
  NodeType childNode;
  NodeType *childPtr = &childNode;
  int index;

  if (!parentNode || !childName ||
      !IsBranch(parentNode->type))
    return NULL;
  
  childNode.name = InstanceString(childName);
 
  index = ArraySearch(parentNode->data.branch_val, &childPtr,
		      NodeNameCompFn, 0, 0);
  
  if (index == -1)
    return NULL;
  else 
    {
      NodeType **ptr = (NodeType**)
	ArrayNth(parentNode->data.branch_val, index);
      assert(ptr != NULL);
      
      return *ptr;
    }
}

unsigned NumChildren(NodeType *parentNode)
{
  if (!IsBranch(parentNode->type)) 
    return 0;
  else
    return ArrayLength(parentNode->data.branch_val);
}

int RemoveNode(NodeType *fromNode, NodeType *node)
{
  int nodeIndex;
  NodeType *foundNode, **foundptr;
  
  if (!fromNode || !node) 
    return 0;
  
  // This is an error, return NULL
  if (!IsBranch(fromNode->type))
    return 0;
  
  // Get the node and its index
  nodeIndex = ArraySearch(fromNode->data.branch_val, &node,
			  PointerCompFn, 0, 0);
  
  if (nodeIndex == -1)
    return 0; // not found
  else
    foundptr = (NodeType**)ArrayNth(fromNode->data.branch_val, nodeIndex);
  
  // Set the return node and decrement its reference count
  foundNode = *foundptr;
 
  // Delete the array space -- does not deallocate node memory
  ArrayRemoveAt(fromNode->data.branch_val, nodeIndex);
  return 1;
}

NodeType *CopyNode(NodeType *src)
{
  NodeType *ret;
  if (!src)
    return NULL;
  
  // Copy leaf node case.
  if (!IsBranch(src->type))
    {
      ret = CreateTypedNode(src->name, src->type, &src->data);
     
      switch (GetFormat(src->type))
	{
	case T_SYMLINK :
	case T_STRING : 
	  if (src->data.string_val)
	    {
	      ret->data.string_val = strdup(src->data.string_val);
	    }
	  break;
	case T_NETSOCK:
	  ret->data.netsock_val = malloc(sizeof(NetSocketLocation));
	  memcpy(ret->data.netsock_val, src->data.netsock_val,
		 sizeof(NetSocketLocation));
	  break;
	case T_DOUBLE:
	  if (src->data.doubleptr_val)
	    {
	      ret->data.doubleptr_val = malloc(sizeof(double));
	      *ret->data.doubleptr_val = *src->data.doubleptr_val;
	    }
	  break;
	case T_TIMEVAL:
	  if (src->data.timeval_val)
	    {
	      ret->data.timeval_val = malloc(sizeof(struct timeval));
	      *ret->data.timeval_val = *src->data.timeval_val;
	    }
	}
    }
  else
    {
      DArray copyNodes, nodes = src->data.branch_val;
      int i, len = ArrayLength(nodes);
      
      ret = CreateBranchNode(src->name, src->type);
      
      copyNodes = ret->data.branch_val;
      for (i = 0; i < len; i++)
	{
	  NodeType *newChild = CopyNode(*((NodeType**)ArrayNth(nodes, i)));
	  ArrayInsertAt(copyNodes, &newChild, i);
	}
    }
  return ret;
}	

static int TraverseTreeRec(NodeType *node, NodeType *parent, void *data,
			   TravTreeFN pre, TravTreeFN post, char *UNL)
{
  if (pre)
    {
      int preRet = 1;
      if (!(preRet = (*pre)(data, node, parent, UNL)))
	return 0;
      if (preRet == 2)
	return 1;
    }
  if (IsBranch(node->type))
    {
      DArray nodes = node->data.branch_val;
      int index = 0;
      NodeType *child;
      while (index < ArrayLength(nodes))
	{
	  char *newUNL = NULL;
	  child = *((NodeType **)ArrayNth(node->data.branch_val, index));
	  if (UNL)
	    newUNL = AppendUNLPath(UNL, child->name);
	  if (!TraverseTreeRec(child, node, data, pre, post, newUNL))
	    return 0;
	  if (index < ArrayLength(nodes) &&
	      child == *((NodeType **)ArrayNth(nodes, index)))
	    index++;
	}
    }
  if (post)
    if (!(*post)(data, node, parent, UNL))
      return 0;
  return 1;
}

int TraverseTree(NodeType *node, NodeType *parent, void *data,
		 TravTreeFN pre, TravTreeFN post,
		 char *UNL)
{
  return TraverseTreeRec(node, parent, data, pre, post, UNL);
}

char *TypeName(TypeID type)
{
  switch (type)
    {
    case T_EMPTY:
      return "empty";
    case T_INT8:
      return "byte";
    case T_INT16:
      return "short";
    case T_INT32:
      return "long";
    case T_FLOAT:
      return "float";
    case T_BOOL:
      return "boolean";
    case T_HANDLE:
      return "handle";
    case T_TYPE:
      return "type";
    case T_STRING:
      return "string";
    case T_DOUBLE:
      return "double";
    case T_SYMLINK:
      return "symlink";
    case T_NETSOCK:
      return "netsock";
    case T_SIGNAL:
      return "signal";
    case T_SERVER:
      return "server";
    case T_BRANCH:
      return "branch";
    case T_DISTINCT:
      return "distinct";
    case T_VECTOR:
      return "vector";
    case T_MATRIX:
      return "matrix";
    case T_TIMEVAL:
      return "time";
    case T_CHARACTER:
      return "character";
    default:
      return "(other)";
    }
}

TypeID TypeByName(char *typename)
{
  if (!strncmp(typename, "T_", 2))
    typename += 2;
  if (!strcasecmp(typename, "STRING"))
    return T_STRING;
  if (!strcasecmp(typename, "INT"))
    return T_INT32;
  if (!strcasecmp(typename, "FLOAT"))
    return T_FLOAT;
  if (!strcasecmp(typename, "DOUBLE"))
    return T_DOUBLE;
  if (!strcasecmp(typename, "DOUBLEPTR"))
    return T_DOUBLE;
  if (!strcasecmp(typename, "SYMLINK"))
    return T_SYMLINK;
  if (!strcasecmp(typename, "DISTINCT"))
    return T_DISTINCT;
  if (!strcasecmp(typename, "BRANCH"))
    return T_BRANCH;
  if (!strcasecmp(typename, "SERVER"))
    return T_SERVER;
  if (!strcasecmp(typename, "VECTOR"))
    return T_VECTOR;
  if (!strcasecmp(typename, "MATRIX"))
    return T_MATRIX;
  if (!strcasecmp(typename, "CHARACTER"))
    return T_CHARACTER;
  return T_NONE;
}

void PrintValue(NodeType *node, FILE *file)
{
  unsigned int ip;
  int i, j;

  if (!file)
    file = stdout;

  if (!node) 
    {
      fprintf(file, "\n");
      return;
    }

  if (IsBranch(node->type))
    {
      unsigned index;
      NodeType *childNode;
      int first = 1;
      for (index = 0; (childNode = BranchIndex(node, index)); index++)
	{
	  if (first)
	    first = 0;
	  else
	    fprintf(file, " ");
	  fprintf(file, "%s", childNode->name);
	}
      fprintf(file, "\n");
    }
  else
    {
      switch (node->type)
	{
	case T_EMPTY:
	  fprintf(file, "\n");
	  break;
	case T_INT8:
	case T_INT16:
	case T_INT32:
	  fprintf(file, "%d\n", node->data.int_val);
	  break;
	case T_CHARACTER:
	  fprintf(file, "%c\n", (char) node->data.int_val);
	  break;
	case T_TYPE:
	  fprintf(file, "%8x\n", node->data.type_val);
	  break;
	case T_FLOAT:
	  fprintf(file, "%f\n", node->data.float_val);
	  break;
	case T_BOOL:  
	  fprintf(file, "%s\n", (node->data.int_val) ? "True" : "False");
	  break;
	case T_STRING:
	  fprintf(file, "%s\n", node->data.string_val);
	  break;
	case T_DOUBLE:
	  fprintf(file, "%g\n", *node->data.doubleptr_val);
	  break;
	case T_TIMEVAL:
	  fprintf(file, "%ld.%06ld\n", 
		  node->data.timeval_val->tv_sec,
		  node->data.timeval_val->tv_usec);
	  break;
	case T_SYMLINK:
	  fprintf(file, "->%s\n", node->data.string_val);
	  break;
	case T_NETSOCK:
	case T_SIGNAL:
	case T_SERVER:
	  ip = ntohl(node->data.netsock_val->ip);
	  fprintf(file, "%d.%d.%d.%d:%d\n",
		  ip >> 24, (ip >> 16) & 255, (ip >> 8) & 255, ip & 255,
		  ntohs(node->data.netsock_val->port));
	  break;
	case T_VECTOR:
	  switch ((node->type >> 12) & 0xF)
	    {
	    case 1:
	      fprintf(file, "[");
	      for (i = 0; i < node->data.vector_val->dimension; i++) 
		{
		  fprintf(file, i ? " %g" : "%g",
			  VectorRef(node->data.vector_val, i));
		}
	      fprintf(file, "]\n");
	      break;
	    }
	  break;
	case T_MATRIX:
	  switch ((node->type >> 12) & 0xF)
	    {
	    case 2:
	      fprintf(file, "[");
	      for (i = 0; i < node->data.matrix_val->rows; i++) 
		{
		  fprintf(file, i ? " [" : "[");
		  for (j = 0; j < node->data.matrix_val->cols; j++) 
		    {
		      fprintf(file, j ? " %g" : "%g",
			      MatrixRef(node->data.matrix_val, i, j));
		    }
		  fprintf(file, "]");
		}
	      fprintf(file, "]\n");
	      break;
	    }
	  break;
	default:
	  fprintf(file, "%d\n", node->data.int_val);
	  break;
	}
    }
}

typedef struct 
{
  int level;
  FILE *file;
  int (*test)(NodeType *);
} PrintType;

static void PrintIndent(PrintType *print) 
{
  int i;
  
  for (i = 0; i < print->level; i++)
    {
      fprintf(print->file, "   ");
    }
}

static int PrintNodePre(PrintType *printData, NodeType *node, 
			NodeType *parent, char *UNL)
{
  if (printData->test && !(*printData->test)(node))
    return 2;
  // Print the common fields of a node
  PrintIndent(printData);
  fprintf(printData->file, "Name: \"%s\", TypeID: %8x\n", 
	  node->name, node->type);

  printData->level++; // Increase for the data

  if (!IsBranch(node->type))
    {
      PrintIndent(printData);
      fprintf(printData->file, "%s: ", TypeName(node->type));
      PrintValue(node, printData->file);
    }
  return 1;
}

static int PrintNodePost(PrintType *printData, NodeType *node, 
			 NodeType *parent, char *UNL)
{
  printData->level--;
  return 1;
}

void PrintNode(NodeType *node, FILE *file)
{
  PrintNodeFilter(node, file, NULL);
}

void PrintNodeFilter(NodeType *node, FILE *file, int (*test)(NodeType *))
{  
  PrintType dPrintData;

  if (!file)
    file = stdout;
  dPrintData.file = file;
  dPrintData.test = test;
  dPrintData.level = 0;

  if (!node) 
    {
      fprintf(file, "(null node)\n");
      return;
    }

  TraverseTree(node, NULL, &dPrintData, (TravTreeFN) PrintNodePre, 
	       (TravTreeFN) PrintNodePost, NULL);
}

NodeType *BranchIndex(NodeType *branch, unsigned index)
{
  if (index >= NumChildren(branch))
    return NULL;
  return *((NodeType**)ArrayNth(branch->data.branch_val, index));
}
