/*************************************************
 * File: journal.c
 * Birth Date: Dec 03, 2003
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2003 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *************************************************/

#include <enchantment/types.h>
#include <stdlib.h>
#include <enchantment/node.h>
#include "ipc.h"
#include "socket.h"
#include "tree-manip.h"
#include "logging.h"
#include <string.h>
#include <fcntl.h>
#include <errno.h>

int ignore_state = 0;
Socket *g_stateSock = NULL;

#define WHITEBOARD_FILE "whiteboard"

int 
HardPublish(NodeType *node, char *UNL)
{
  int size = 0;
  char *data = NULL;

  if (ignore_state) // we're still playing the journal
    return 1;

  plogf("Hard publishing to %s:\n", UNL);
  logNode(node);

  if (!g_stateSock)
    {
      printf("No state supported\n");
      return 0;
    }
  if (!SocketSendRecord(g_stateSock, UNL, strlen(UNL) + 1))
    {
      printf("write failed 1\n");
      return 0;
    }
  if (!node)
    {
      if (!SocketSendRecord(g_stateSock, data, size))
	{
	  printf("write failed 2\n");
	  return 0;
	}
      //SockSendFlush(g_stateSock);
      return 1;
    }
  WriteNode(g_stateSock, node, NULL, NULL);
  return 1;
}

static int RewriteJournalPre(void *data, NodeType *node, NodeType *parent, 
			     char *UNL)
{
  if (SchemaOf(node)->hardNode)
    {
      HardPublish(node, UNL);
      return 2;
    }
  else
    return 1;
}

int RewriteJournal()
{
  return TraverseTree(g_globalNode, NULL, NULL, 
		      (TravTreeFN) RewriteJournalPre, NULL, "");
}

int
PlayJournalEntry(Socket *sock)
{
  char *UNL;
  int size;
  NodeType *node = NULL;
  if (!SocketRecvRecord(sock, &UNL, &size))
    {
      if (SocketError(sock))
	{
	  printf("Error at start of record: %s\n", 
		 strerror(SocketError(sock)));
	}
      return 0;
    }
  UNL = strdup(UNL);
  ReadNode(sock, &node, NULL, NULL);
  /*
  printf("Read a record for %s\n", UNL);
  PrintNode(node, NULL);
  */
  PublishLocal(UNL, node, 1);
  free(UNL);
  //printf("Journal entry successfully played\n");
  return 1;
}

int LoadStateFile(int quiet, char *filename)
{
  int fd; 
  int newfd;
  int i;
  char *tmpname;
  char *statename;
  struct flock lock;
  Socket *readSock;

  statename = malloc(strlen(filename) + strlen(WHITEBOARD_FILE) + 2);
  strcpy(statename, filename);
  strcpy(statename + strlen(filename), "/" WHITEBOARD_FILE);
  if (!quiet)
    printf("Using %s as state file\n", statename);
  fd = open(statename, O_RDONLY);
  if (fd == -1 && errno != ENOENT)
    {
      if (SERVER_ERROR) printf("Could not open %s as state file: %s\n",
			       statename, strerror(errno));
      free(statename);
      return 0;
    }
  lock.l_type = F_WRLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start = 0;
  lock.l_len = 0;
  lock.l_pid = 0;
  if (fd >= 0)
    {
      if (fcntl(fd, F_GETLK, &lock) == -1)
	{
	  if (errno == ENOLCK)
	    {
	      printf("No locks supported/available; continuing without lock.\n");
	    }
	  else 
	    {
	      if (SERVER_ERROR) printf("%s has problem: %s\n",
				       statename, strerror(errno));
	      close(fd);
	      free(statename);
	      return 0;
	    }
	}
      if (lock.l_pid)
	{
	  if (SERVER_ERROR) printf("%s is locked by another server instance.\n",
				   statename);
	  close(fd);
	  free(statename);
	  return 0;
	}
    }
  tmpname = malloc(strlen(statename) + 5);
  strcpy(tmpname, statename);
  strcpy(tmpname + strlen(statename), "-new");
  newfd = open(tmpname, O_EXCL | O_CREAT | O_WRONLY, 0444);
  if (newfd == -1)
    {
      if (SERVER_ERROR) printf("Could not open %s as new state file: %s\n",
			       tmpname, strerror(errno));
      free(tmpname);
      if (fd != -1)
	close(fd);
      free(statename);
      return 0;
    }
  lock.l_type = F_WRLCK;
  if (fcntl(newfd, F_SETLK, &lock) == -1)
    {
      if (errno == ENOLCK)
	{
	  /* We presumably know... */
	}
      else
	{
	  if (SERVER_ERROR) printf("Could not lock new state file: %s\n",
				   strerror(errno));
	  if (fd != -1)
	    close(fd);
	  close(newfd);
	  remove(tmpname);
	  free(tmpname);
	  free(statename);
	  return 0;
	}
    }
  /* Everything is now set up: fd contains the old journal, which is to be
   * read; newfd is (barring NFS issues and races) exclusively the new
   * journal file.
   */
  
  /* Read in the old journal, if there is one. */

  if (fd >= 0)
    {
      readSock = SocketFile(fd);
      i = 0;
      // We're going to write the tree after the journal is played.
      ignore_state = 1;
      while (PlayJournalEntry(readSock))
	i++;
      ignore_state = 0;
      if (!quiet)
	printf("Played %d record(s).\n", i);
      close(fd);
      SocketClose(readSock);
      //PrintNode(g_globalNode, NULL);
    }
  else
    {
      printf("New whiteboard state.\n");
    }

  /* Write out the state. */
  //printf("fd %d for new state\n", newfd);
  g_stateSock = SocketFile(newfd);
  //PrintNode(g_globalNode, NULL);
  RewriteJournal();
  
  /* New journal is now complete. Put it in place. */
  if (rename(tmpname, statename))
    {
      printf("Error moving %s to %s: %s\n", tmpname, statename, 
	     strerror(errno));
    }
  free(tmpname);
  free(statename);
  return 1;
}
